; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Initialization
; ======================================================================

; Set up the HUD camera and default settings
Function nS_Initialize()
	nS_Assert(Not nS_Camera, "Initialization has already been done")
	nS_AspectRatio# = Float(GraphicsHeight()) / Float(GraphicsWidth())
	nS_Camera = CreateCamera()
	CameraClsMode nS_Camera, False, True
	CameraRange nS_Camera, 1, 101
	nS_ZBuffer = 0
	CameraProjMode nS_Camera, 2
	PositionEntity nS_Camera, 0, -1000, 0
	nS_pvtTopLeft = CreatePivot(nS_Camera)
	PositionEntity nS_pvtTopLeft, -1, nS_AspectRatio#, 100
	nS_Scale = Float(2) / Float(GraphicsWidth())
	ScaleEntity nS_pvtTopLeft, nS_Scale, -nS_Scale, -nS_Scale
	nS_imgBlank = nS_CreateImage(2, 2, 1, 0)
	nS_DefaultFX = nS_FX_FULLBRIGHT
	nS_AnimationTween = 1
	nS_ShapesMesh = CreateMesh(nS_pvtTopLeft)
	nS_ShapesSurface = CreateSurface(nS_ShapesMesh)
	AddVertex(nS_ShapesSurface, 0, 0, 0)
	EntityFX nS_ShapesMesh, 35
End Function

; ======================================================================
; User Functions
; ======================================================================

; Changes the viewport to display the sprites
Function nS_Viewport(iX%, iY%, iW%, iH%, bUseOffset% = False)
	CameraViewport nS_Camera, iX, iY, iW, iH
	CameraZoom nS_Camera, Float(GraphicsWidth())/Float(iW)
	nS_Origin(GraphicsWidth()/2-iW/2-iX*(bUseOffset=False), GraphicsHeight()/2-iH/2-iY*(bUseOffset=False))
End Function

; Sets the origin to draw sprites from
Function nS_Origin(iX%, iY%)
	nS_OriginX = iX
	nS_OriginY = iY
End Function

; Clears the screen
Function nS_Cls(bClear% = True)
	Local img.nS_Image
	Local i%, j%
	If bClear Then
		Cls
	End If
	; Clear images
	For img = Each nS_Image
		For i = 0 To img\instanceMax
			If img\instanceUsed[i] Then
				img\instanceUsed[i] = False
				For j = 0 To 3
					VertexCoords img\surface, img\instanceID[i]+j, 0, 0, 0
				Next
			End If
		Next
	Next
	; Clear basic shapes
	For i = 1 To nS_ShapesCount
		nS_ShapesUsed[i] = False
		For j = nS_ShapesStart[i] To nS_ShapesStart[i] + nS_ShapesVertices[i] - 1
			VertexCoords nS_ShapesSurface, j, 0, 0, 0
		Next
	Next
	; Clear buffers
	nS_SpriteCount = 0
	nS_ZBuffer = 0
End Function

; Clears everything
Function nS_ClearWorld()
	ClearWorld()
	For img.nS_Image = Each nS_Image
		Delete img
	Next
	For fnt.nS_Font = Each nS_Font
		Delete fnt
	Next
	For col.nS_Collision = Each nS_Collision
		Delete col
	Next
	For lgt.nS_Light = Each nS_Light
		Delete lgt
	Next
	For part.nS_Particle = Each nS_Particle
		Delete part
	Next
	For emt.nS_Emitter = Each nS_Emitter
		Delete emt
	Next
	For evt.nS_Event = Each nS_Event
		Delete evt
	Next
	For tmr.nS_Timer = Each nS_Timer
		Delete tmr
	Next
	nS_Camera = 0
	; Clear basic shapes
	For i = 1 To nS_ShapesCount
		nS_ShapesUsed[i] = False
		nS_ShapesStart[i] = 0
		nS_ShapesVertices[i] = 0
		nS_ShapesType[i] = 0
	Next
	; Clear buffers
	nS_ShapesCount = 0
	nS_SpriteCount = 0
	nS_ZBuffer = 0
End Function

; Changes the colors
Function nS_Color(iRed%, iGreen%, iBlue%, fAlpha# = 1, bChange2D% = True)
	nS_ColorRed = iRed
	nS_ColorGreen = iGreen
	nS_ColorBlue = iBlue
	nS_ColorAlpha = fAlpha
	If bChange2D Then
		Color iRed, iGreen, iBlue
	End If
End Function

; Draws a rectangle on screen
Function nS_Rect(fX#, fY#, fWidth#, fHeight#, bSolid% = True, fBorderThickness# = 1.0)
	Local p%, v%, fThick#
	fX = fX - 0.5 + nS_OriginX
	fY = fY - 0.5 + nS_OriginY
	If bSolid Then
		p = nS_FindShape(nS_SHAPE_FILLRECT_LINE_OR_POINT, 4)
		v = nS_ShapesStart[p]
		VertexCoords nS_ShapesSurface, v, fX, fY, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+1, fX, fY+fHeight, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+2, fX+fWidth, fY, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+3, fX+fWidth, fY+fHeight, nS_ZBuffer
	Else
		fThick = fBorderThickness / 2.0
		p = nS_FindShape(nS_SHAPE_HOLLOWRECT, 8)
		v = nS_ShapesStart[p]
		VertexCoords nS_ShapesSurface, v, fX-fThick, fY-fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+1, fX-fThick, fY+fHeight+fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+2, fX+fWidth+fThick, fY-fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+3, fX+fWidth+fThick, fY+fHeight+fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+4, fX+fThick, fY+fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+5, fX+fThick, fY+fHeight-fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+6, fX+fWidth-fThick, fY+fThick, nS_ZBuffer
		VertexCoords nS_ShapesSurface, v+7, fX+fWidth-fThick, fY+fHeight-fThick, nS_ZBuffer
	End If
	; Positioning
	PositionEntity nS_ShapesMesh, 0, 0, nS_ZBuffer
	; ZBuffer update
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function

; Draws a line
Function nS_Line(fX1#, fY1#, fX2#, fY2#, fThick# = 1.0)
	Local p%, v%, fX#, fY#, fP#, dx#, dy#
	dx = fX1 - fX2
	dy = fY1 - fY2
	p = nS_FindShape(nS_SHAPE_FILLRECT_LINE_OR_POINT, 4)
	v = nS_ShapesStart[p]
	fP = 90 - (VectorYaw(dx, 0, dy) - 90)
	fThick = fThick / 2.0
	fX = Cos(fP) * fThick
	fY = Sin(fP) * fThick
	fX1 = fX1 - 0.5 + nS_OriginX
	fY1 = fY1 - 0.5 + nS_OriginY
	fX2 = fX2 - 0.5 + nS_OriginX
	fY2 = fY2 - 0.5 + nS_OriginY
	VertexCoords nS_ShapesSurface, v, fX1+fX, fY1-fY, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+1, fX1-fX, fY1+fY, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+2, fX2+fX, fY2-fY, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+3, fX2-fX, fY2+fY, nS_ZBuffer
	; Positioning
	PositionEntity nS_ShapesMesh, 0, 0, nS_ZBuffer
	; ZBuffer update
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function

; Draws an oval
Function nS_Oval(fX#, fY#, fWidth#, fHeight#, bSolid% = True, fBorderThickness# = 1.0, iSides% = 32, fAngle# = 0.0)
	Local p%, v%, fThick#, fA#
	fWidth = fWidth / 2.0
	fHeight = fHeight / 2.0
	fX = fX - 0.5 + fWidth + nS_OriginX
	fY = fY - 0.5 + fHeight + nS_OriginY
	fA = 360.0 / Float(iSides)
	If bSolid Then
		p = nS_FindShape(nS_SHAPE_FILLOVAL, iSides+1)
		v = nS_ShapesStart[p]
		VertexCoords nS_ShapesSurface, v, fX, fY, nS_ZBuffer
		For i = 1 To iSides
			VertexCoords nS_ShapesSurface, v+i, fX+fWidth*Cos(fAngle), fY+fHeight*Sin(fAngle), nS_ZBuffer
			fAngle = fAngle - fA
		Next
	Else
		fThick = fBorderThickness / 2.0 + 1.0/Float(iSides-1)
		p = nS_FindShape(nS_SHAPE_HOLLOWOVAL, iSides*2)
		v = nS_ShapesStart[p]
		For i = 0 To iSides-1
			VertexCoords nS_ShapesSurface, v+i*2, fX+(fWidth-fThick)*Cos(fAngle), fY+(fHeight-fThick)*Sin(fAngle), nS_ZBuffer
			VertexCoords nS_ShapesSurface, v+i*2+1, fX+(fWidth+fThick)*Cos(fAngle), fY+(fHeight+fThick)*Sin(fAngle), nS_ZBuffer
			fAngle = fAngle - fA
		Next
	End If
	; Positioning
	PositionEntity nS_ShapesMesh, 0, 0, nS_ZBuffer
	; ZBuffer update
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function

; Draws a pixel on screen
Function nS_Plot(fX#, fY#)
	Local p%, v%
	fX = fX - 0.5 + nS_OriginX
	fY = fY - 0.5 + nS_OriginY
	p = nS_FindShape(nS_SHAPE_FILLRECT_LINE_OR_POINT, 4)
	v = nS_ShapesStart[p]
	VertexCoords nS_ShapesSurface, v, fX, fY, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+1, fX, fY+1, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+2, fX+1, fY, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+3, fX+1, fY+1, nS_ZBuffer
	; Positioning
	PositionEntity nS_ShapesMesh, 0, 0, nS_ZBuffer
	; ZBuffer update
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function

; Draws a triangle
Function nS_Triangle(fX1#, fY1#, fX2#, fY2#, fX3#, fY3#)
	Local p%, v%, fThick#, fOX#, fOY#, l#
	fX1 = fX1 - 0.5
	fY1 = fY1 - 0.5
	fX2 = fX2 - 0.5
	fY2 = fY2 - 0.5
	fX3 = fX3 - 0.5
	fY3 = fY3 - 0.5
	p = nS_FindShape(nS_SHAPE_TRI, 3)
	v = nS_ShapesStart[p]
	VertexCoords nS_ShapesSurface, v, fX1, fY1, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+1, fX2, fY2, nS_ZBuffer
	VertexCoords nS_ShapesSurface, v+2, fX3, fY3, nS_ZBuffer
	; Positioning
	PositionEntity nS_ShapesMesh, 0, 0, nS_ZBuffer
	; ZBuffer update
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function