; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Globals and Constants
; ======================================================================
Const nS_IMGSIZE = 128
Const nS_MAXCHARS = 255

Global nS_CurrentFont

; ======================================================================
; Different objects
; ======================================================================

; Font Object
Type nS_Font
	Field sprFont				; Sprite of the characters
	Field sprASCII[nS_MAXCHARS] ; Individual character sprites
	Field iSpacing				; Spacing between characters
	Field iLarge				; Largest Character Width
End Type

; ======================================================================
; Functions
; ======================================================================

; Creates a new font from parameters
Function nS_CreateFont%(Font$, Size%, Bold%, Italic%, Underline%, ColorLevel%, OutlineColorLevel%, OutlineWeight%, Spacing%, iFirstChar% = 0, iLastChar% = 255, iTexSize% = -1)
	Local newfont.nS_Font = New nS_Font
	Local fnt = LoadFont(Font$, Size, Bold, Italic, Underline)
	Local img = CreateImage(nS_IMGSIZE, nS_IMGSIZE), imgASCII
	Local c, x, y, i, j
	Local minY, minX, w, h, p
	Local bAllBlack
	Local texX, texY, iLarge, maxX, iTallest

	; Predict font size
	SetFont fnt
	iLarge = StringWidth("W") + OutlineWeight*2
	If iTexSize = -1 Then
		maxX = 2 ^ Ceil(nS_Log2((Sqr(iLarge * (iLastChar - iFirstChar)))^1.4))
	Else
		maxX = iTexSize
	End If

	; Create the font
	newfont\sprFont = nS_CreateImage(maxX, maxX)
	LockBuffer TextureBuffer(nS_TextureHandle(newfont\sprFont))
	newfont\iSpacing = spacing
	newfont\iLarge = iLarge
	
	; Create character by character
	SetBuffer ImageBuffer(img)
	minY = 16 - OutlineWeight
	ClsColor 0, 0, 0
	For c = iFirstChar To iLastChar
		Cls
		Color OutlineColorLevel, OutlineColorLevel, OutlineColorLevel
		For x = 0 To OutlineWeight * 2
			For y = 0 To OutlineWeight * 2
				Text x + 16, y + 16, Chr(c)
			Next	
		Next 
		Color ColorLevel, ColorLevel, ColorLevel
		Text OutlineWeight + 16, OutlineWeight + 16, Chr(c)
		minX = -1
		h = -1
		For i = 0 To nS_IMGSIZE
			bAllBlack = True
			For j = minY To nS_IMGSIZE
				p = ReadPixel(i, j)
				If p <> 0 And p <> -16777216 Then
					If j - minY + 1 > h Then
						h = j - minY + 1
					End If
					bAllBlack = False
				End If
			Next
			If minX = -1 Then
				If Not bAllBlack Then
					minX = i
				End If
			Else
				If bAllBlack Then
					w = i - minX
					i = nS_IMGSIZE
				End If
			End If
		Next
		If h = -1 Then
			h = 1
		End If
		If w = -1 Then
			w = 1
		End If
		If c = 32 Then
			w = Size / 3
		End If
		If h > iTallest Then
			iTallest = h
		End If

		; Put the character into the texture
		LockBuffer ImageBuffer(img)
			For x = 0 To w - 1
				For y = 0 To h - 1
					nS_Assert(y + texY <= maxX, "Algorithm determined an incorrect texture size. Please specify one greater than " + maxX + ".")
					p = ReadPixelFast(x + minX, y + minY, ImageBuffer(img))
					If p = -16777216 Then
						p = 0
					End If
					WritePixelFast x + texX, y + texY, p, TextureBuffer(nS_TextureHandle(newfont\sprFont))
				Next
			Next
		UnlockBuffer ImageBuffer(img)

		; Create individual sprite
		newfont\sprASCII[c] = nS_CreateSubImage(newfont\sprFont, texX, texY, w, h)

		; Change line
		texX = texX + w
		If texX + iLarge > maxX Then
			texX = 0
			texY = texY + iTallest
			iTallest = 0
		End If
		
	Next
	UnlockBuffer TextureBuffer(nS_TextureHandle(newfont\sprFont))
	SetBuffer BackBuffer()
	FreeFont fnt
	FreeImage img
	Return Handle(newfont)
End Function

; Saves font to a file
Function nS_SaveFont(Font%, sFile$)
	Local oFont.nS_Font = Object.nS_Font(Font)
	Local fs, c, x, y, p, w, h
	fs = WriteFile(sFile$)
		; Version 1.1.0 font saver
		WriteInt fs, 110
		WriteInt fs, oFont\iSpacing
		WriteInt fs, oFont\iLarge
		WriteInt fs, nS_ImageWidth(oFont\sprFont)
		LockBuffer(nS_ImageBuffer(oFont\sprFont))
		For y = 0 To nS_ImageWidth(oFont\sprFont) - 1
			For x = 0 To nS_ImageWidth(oFont\sprFont) - 1
				p = ReadPixelFast(x, y, nS_ImageBuffer(oFont\sprFont))
				If p = -16777216 Then
					p = 0
				End If
				WriteInt fs, p
			Next
		Next
		UnlockBuffer(nS_ImageBuffer(oFont\sprFont))
		For c = 0 To 255
			If oFont\sprASCII[c] <> 0 Then
				WriteInt fs, c
				w = nS_ImageWidth(oFont\sprASCII[c])
				h = nS_ImageHeight(oFont\sprASCII[c])
				WriteInt fs, w
				WriteInt fs, h
			End If
		Next
	CloseFile fs
End Function

; Loads font from a file
Function nS_LoadFont%(sFile$)
	Local newfont.nS_Font = New nS_Font
	Local x, y, w, h, p, c, maxX, texX, texY, iTallest
	Local header, version
	Local fs = ReadFile(sFile$)
		version = ReadInt(fs)
		; Version 1.1.0 font loader
		nS_Assert(version <= 110, "Version not compatible with this font version!")
		If version <= 110 Then
			newfont\iSpacing = ReadInt(fs)
			newfont\iLarge = ReadInt(fs)
			maxX = ReadInt(fs)
			newfont\sprFont = nS_CreateImage(maxX, maxX)
			LockBuffer(nS_ImageBuffer(newfont\sprFont))
			For y = 0 To maxX - 1
				For x = 0 To maxX - 1
					p = ReadInt(fs)
					WritePixelFast x, y, p, nS_ImageBuffer(newfont\sprFont)
				Next
			Next
			UnlockBuffer(nS_ImageBuffer(newfont\sprFont))
			While Not Eof(fs)
				c = ReadInt(fs)
				w = ReadInt(fs)
				h = ReadInt(fs)
				newfont\sprASCII[c] = nS_CreateSubImage(newfont\sprFont, texX, texY, w, h)
				; Change line
				texX = texX + w
				If h > iTallest Then
					iTallest = h
				End If
				If texX + newfont\iLarge > maxX Then
					texX = 0
					texY = texY + iTallest
					iTallest = 0
				End If
			Wend
		End If
	CloseFile fs
	Return Handle(newfont)
End Function

; Frees font from memory
Function nS_FreeFont(hFont%)
	Local oFont.nS_Font = Object.nS_Font(hFont)
	Local i
	nS_FreeImage(oFont\sprFont)
	Delete oFont
End Function

; Set font color
Function nS_FontColor(font, red, green, blue, alpha# = 1)
	Local oFont.nS_Font = Object.nS_Font(Font)
	nS_ImageColor(oFont\sprFont, red, green, blue)
	nS_ImageAlpha(oFont\sprFont, alpha)
End Function

; Set font to use
Function nS_SetFont(font%)
	nS_CurrentFont = font
End Function

; Displays a string on screen
Function nS_Text(x#, y#, sstring$, centre_x% = False, centre_y% = False, font% = -1, width_limit# = -1, line_height# = -1, trimword% = True)
	Local w# = 0, bi%
	Local oFont.nS_Font
	If font = -1 Then
		font = nS_CurrentFont
	End If
	oFont = Object.nS_Font(font)
	; No width limit
	If width_limit = -1 Then
		If centre_x Then
			x = x - nS_StringWidth(sstring, font) / 2
		End If
		If centre_y Then
			y = y - nS_StringHeight(sstring, font) / 2
		End If
		For i = 1 To Len(sstring)
			c = Asc(Mid(sstring, i, 1))
			nS_Assert(oFont\sprASCII[c] <> 0, "Character " + c + "(" + Chr(c) + ") not in font!")
			nS_DrawImage oFont\sprASCII[c], x, y
			x = x + nS_ImageWidth(oFont\sprASCII[c]) + oFont\iSpacing
		Next
	; Width limit
	Else
		If line_height = -1 Then
			line_height = nS_StringHeight(sstring, font)
		End If
		For i = 1 To Len(sstring)
			c = Asc(Mid(sstring, i, 1))
			nS_Assert(oFont\sprASCII[c] <> 0, "Character " + c + "(" + Chr(c) + ") not in font!")
			w = w + nS_ImageWidth(oFont\sprASCII[c]) + oFont\iSpacing
			If w > width_limit + oFont\iSpacing Then
				i = i - 1
				If trimword Then
					bi = i
					While Right(Left(sstring, i), 1) <> " " And i > 1
						i = i - 1
					Wend
					If i = 1 Then
						i = bi
					End If
					If i = 0 Then
						i = 1
					End If
					nS_Text x, y, Trim(Left(sstring, i)), centre_x, centre_y, font
					nS_Text x, y + line_height, Trim(Mid(sstring, i+1)), centre_x, centre_y, font, width_limit, line_height
				Else
					If i = 0 Then
						i = 1
					End If
					nS_Text x, y, Left(sstring, i), centre_x, centre_y, font
					nS_Text x, y + line_height, Mid(sstring, i+1), centre_x, centre_y, font, width_limit, line_height
				End If
				Return
			End If
			If c = 10 Or c = 13 Then
				If trimword Then
					nS_Text x, y, Trim(Left(sstring, i-1)), centre_x, centre_y, font
					nS_Text x, y + line_height, Trim(Mid(sstring, i+1)), centre_x, centre_y, font, width_limit, line_height
				Else
					nS_Text x, y, Left(sstring, i-1), centre_x, centre_y, font
					nS_Text x, y + line_height, Mid(sstring, i+1), centre_x, centre_y, font, width_limit, line_height
				End If
				Return
			End If
		Next
		If trimword Then
			nS_Text x, y, Trim(sstring), centre_x, centre_y, font
		Else
			nS_Text x, y, Trim(sstring), centre_x, centre_y, font
		End If
	End If
End Function

; Get the width of a string
Function nS_StringWidth%(sstring$, font% = -1)
	Local w = 0
	If font = -1 Then
		font = nS_CurrentFont
	End If
	Local oFont.nS_Font = Object.nS_Font(Font)
	For i = 1 To Len(sstring)
		c = Asc(Mid(sstring, i, 1))
		nS_Assert(oFont\sprASCII[c] <> 0, "Character " + c + "(" + Chr(c) + ") not in font!")
		w = w + nS_ImageWidth(oFont\sprASCII[c]) + oFont\iSpacing
	Next
	If Not sstring = "" Then
		w = w - oFont\iSpacing
	End If
	Return w
End Function

; Get the height of a string
Function nS_StringHeight%(sstring$, font% = -1)
	Local c% = 32, i% = 1
	If font = -1 Then
		font = nS_CurrentFont
	End If
	Local oFont.nS_Font = Object.nS_Font(Font)
	While (c = 32) And (i < Len(sstring))
		c = Asc(Mid(sstring, i, 1))
		i = i + 1
	Wend
	Return nS_ImageHeight(oFont\sprASCII[c])
End Function