; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Globals and Constants
; ======================================================================
Const nS_MAX_COLLISIONS = 16

; ======================================================================
; Different objects
; ======================================================================

; Collision Spheres
Type nS_Collision
	Field iX#			; X Position
	Field iY#			; Y Position
	Field fRadius#		; Radius
	Field iType			; Collision Type
	Field bPoint		; Is it a point
	; Quick Access
	Field fRealX#		; Real X Position
	Field fRealY#		; Real Y Position
	Field fRealRadius#  ; Real Radius
End Type

; ======================================================================
; Functions
; ======================================================================

; Creates a collision sphere or point for an image (returns the ID of the collision sphere)
Function nS_CreateCollision%(Image%, X#, Y#, fRadius#, iType%, bPoint% = False)
	img.nS_Image = Object.nS_Image(Image)
	; Find an open spot
	For i = 0 To nS_MAX_COLLISIONS
		If img\collisionData[i] = Null Then
			img\collisionData[i] = New nS_Collision
			img\collisionData[i]\iX = X
			img\collisionData[i]\iY = Y
			img\collisionData[i]\fRealX = X
			img\collisionData[i]\fRealY = Y
			img\collisionData[i]\iType = iType
			If bPoint Then
				img\collisionData[i]\bPoint = True
			Else
				img\collisionData[i]\fRadius = fRadius
				img\collisionData[i]\fRealRadius = fRadius
			End If
			nS_UpdateCollisions(Image)
			Return i
		End If
	Next
	RuntimeError "No more collisions available for this image"
End Function

; ----------------------------------------------------------------------

; Creates a sphere collision
Function nS_AddCollisionSphere%(Image%, X#, Y#, fRadius#, iType%)
	Return nS_CreateCollision(Image, X, Y, fRadius#, iType, False)
End Function

; ----------------------------------------------------------------------

; Creates a polygon point collision
Function nS_AddCollisionPolyPoint%(Image%, X#, Y#, iType%)
	Return nS_CreateCollision(Image, X, Y, 0, iType, True)
End Function

; ----------------------------------------------------------------------

; Updates the real collision data for quick access
Function nS_UpdateCollisions(Image%)
	img.nS_Image = Object.nS_Image(Image)
	Local fX#, fY#, fRad#, fAngle#, f1#, f2#
	img\fColXMin = 9999999
	img\fColYMin = 9999999
	img\fColXMax = -9999999
	img\fColYMax = -9999999
	; Find a valid collision setting
	For i = 0 To nS_MAX_COLLISIONS
		If img\collisionData[i] <> Null Then
			; Standard data
			fX = img\collisionData[i]\iX
			fY = img\collisionData[i]\iY
			; Scale
			fX = fX * nS_ImageScaleX(Image)
			fY = fY * nS_ImageScaleY(Image)
			; Radius
			If Not img\collisionData[i]\bPoint Then
				fRad = img\collisionData[i]\fRadius
				fRad = fRad * (nS_ImageScaleX(Image) + nS_ImageScaleY(Image)) / 2
				img\collisionData[i]\fRealRadius = fRad
			Else
				fRad = 0
			End If
			; Handle
			fX = fX - nS_ImageXHandle(Image)
			fY = fY - nS_ImageYHandle(Image)
			; Rotation
			fAngle = nS_ImageAngle(Image)
			f1 = Cos(fAngle) * fX - Sin(fAngle) * fY
			f2 = Sin(fAngle) * fX + Cos(fAngle) * fY
			fX = f1
			fY = f2
			; Final Values
			img\collisionData[i]\fRealX = fX
			img\collisionData[i]\fRealY = fY
			; Check for englobing collision
			If fX - fRad < img\fColXMin Then
				img\fColXMin = fX - fRad
			End If
			If fY - fRad < img\fColYMin Then
				img\fColYMin = fY - fRad
			End If
			If fX + fRad > img\fColXMax Then
				img\fColXMax = fX + fRad
			End If
			If fY + fRad > img\fColYMax Then
				img\fColYMax = fY + fRad
			End If
		End If
	Next
End Function

; ----------------------------------------------------------------------

; Draws the collisions on screen
Function nS_DrawCollisions(Image%, X#, Y#)
	Local bFirst = True
	Local lX#, lY#
	img.nS_Image = Object.nS_Image(Image)
	; Draw englobing
	nS_Color 0, 0, 255
	nS_Rect X + img\fColXMin, Y + img\fColYMin, img\fColXMax - img\fColXMin + 1, img\fColYMax - img\fColYMin + 1, False
	; Draw each collision
	For i = 0 To nS_MAX_COLLISIONS
		If img\collisionData[i] <> Null Then
			If img\collisionData[i]\bPoint Then
				If bFirst Then
					bFirst = False
					; First point
					lX = X + img\collisionData[i]\fRealX
					lY = Y + img\collisionData[i]\fRealY
				Else
					nS_Color 255, 255, 0
					; Draw a poly segment
					nS_Line lX, lY, X + img\collisionData[i]\fRealX, Y + img\collisionData[i]\fRealY
					lX = X + img\collisionData[i]\fRealX
					lY = Y + img\collisionData[i]\fRealY
				End If
			Else
				nS_Color 255, 0, 0
				; Draw a sphere
				nS_Oval X + img\collisionData[i]\fRealX - img\collisionData[i]\fRealRadius, Y + img\collisionData[i]\fRealY - img\collisionData[i]\fRealRadius, img\collisionData[i]\fRealRadius * 2, img\collisionData[i]\fRealRadius * 2, False
			End If
		End If
	Next
End Function

; Returns true if the 2 types of primitives have collided
Function nS_Collided%(Image1%, X1#, Y1#, Image2%, X2#, Y2#, Type1% = -1, Type2% = -1, bSkipEnglobing% = False)
	Local dX#, dY#, dSqr#
	Local bFirst1, bFirst2, bFirst = False
	Local lX1#, lY1#
	Local lX2#, lY2#
	Local pX1#, pY1#, pX2#, pY2#, pXT#, pYT#
	Local d1#, d2#
	Local CX1#, CX2#, CY1#, CY2#
	Local tx1#, ty1#, tx2#, ty2#, tx3#, ty3#, tx4#, ty4#
	sprite1.nS_Image = Object.nS_Image(Image1)
	sprite2.nS_Image = Object.nS_Image(Image2)
	; Check englobing collision
	If bSkipEnglobing Then
		tx1 = X1 + sprite1\fColXMin
		ty1 = Y1 + sprite1\fColYMin
		tx2 = X1 + sprite1\fColXMax
		ty2 = Y1 + sprite1\fColYMax
		tx3 = X2 + sprite2\fColXMin
		ty3 = Y2 + sprite2\fColYMin
		tx4 = X2 + sprite2\fColXMax
		ty4 = Y2 + sprite2\fColYMax
		If Not (tx3 > tx1 And ty3 > ty1 And tx3 < tx2 And ty3 < ty2) Then
			If Not (tx4 > tx1 And ty4 > ty1 And tx4 < tx2 And ty4 < ty2) Then
				If Not (tx3 > tx1 And ty4 > ty1 And tx3 < tx2 And ty4 < ty2) Then
					If Not (tx4 > tx1 And ty3 > ty1 And tx4 < tx2 And ty3 < ty2) Then
						Return False
					End If
				End If
			End If
		End If
	End If
	; Check all collisions
	bFirst1 = True
	For i = 0 To nS_MAX_COLLISIONS
		bFirst2 = True		
		For j = 0 To nS_MAX_COLLISIONS
			If sprite1\collisionData[i] <> Null Then
				If sprite2\collisionData[j] <> Null Then
					If sprite1\collisionData[i]\iType = Type1 Or Type1 = -1 Then
						If sprite2\collisionData[j]\iType = Type2 Or Type2 = -1 Then
							; Line versus circle
							If (sprite1\collisionData[i]\bPoint) And (Not sprite2\collisionData[j]\bPoint) Then
								If bFirst1 Then
									bFirst1 = False
								Else
									; Check
									CX1# = sprite1\collisionData[i]\fRealX# + X1
									CY1# = sprite1\collisionData[i]\fRealY# + Y1
									CX2# = sprite2\collisionData[j]\fRealX# + X2
									CY2# = sprite2\collisionData[j]\fRealY# + Y2
									pX1 = lX1 - CX1
									pY1 = lY1 - CY1
									pX2 = CX2 - CX1
									pY2 = CY2 - CY1
									d1 = pX1 * pX2 + pY1 * pY2
									If d1 <= 0 Then
										dX# = CX2 - CX1
										dY# = CY2 - CY1
										dSqr# = dX * dX + dY * dY
									Else
										d2 = pX1 * pX1 + pY1 * pY1
										If (d2 <= d1) Then
											dX# = CX2 - lX1
											dY# = CY2 - lY1
											dSqr# = dX * dX + dY * dY
										Else
											d1 = d1 / d2
											pXT# = CX1 + d1 * pX1
											pYT# = CY1 + d1 * pY1
											dX# = CX2 - pXT
											dY# = CY2 - PYT
											dSqr# = dX * dX + dY * dY
										End If
									End If
									If dSqr# < sprite2\collisionData[j]\fRealRadius * sprite2\collisionData[j]\fRealRadius Then
										Return True
									End If
								End If
							; Circle versus line
							Else If (Not sprite1\collisionData[i]\bPoint) And (sprite2\collisionData[j]\bPoint) Then
								If bFirst2 Then
									bFirst2 = False
								Else
									; Check
									CX1# = sprite2\collisionData[j]\fRealX# + X2
									CY1# = sprite2\collisionData[j]\fRealY# + Y2
									CX2# = sprite1\collisionData[i]\fRealX# + X1
									CY2# = sprite1\collisionData[i]\fRealY# + Y1
									pX1 = lX2 - CX1
									pY1 = lY2 - CY1
									pX2 = CX2 - CX1
									pY2 = CY2 - CY1
									d1 = pX1 * pX2 + pY1 * pY2
									If d1 <= 0 Then
										dX# = CX2 - CX1
										dY# = CY2 - CY1
										dSqr# = dX * dX + dY * dY
									Else
										d2 = pX1 * pX1 + pY1 * pY1
										If (d2 <= d1) Then
											dX# = CX2 - lX1
											dY# = CY2 - lY1
											dSqr# = dX * dX + dY * dY
										Else
											d1 = d1 / d2
											pXT# = CX1 + d1 * pX1
											pYT# = CY1 + d1 * pY1
											dX# = CX2 - pXT
											dY# = CY2 - PYT
											dSqr# = dX * dX + dY * dY
										End If
									End If
									If dSqr# < sprite1\collisionData[i]\fRealRadius * sprite1\collisionData[i]\fRealRadius Then
										Return True
									End If									
								End If
							; Line versus line
							Else If (sprite1\collisionData[i]\bPoint) And (sprite2\collisionData[j]\bPoint) Then
								If bFirst1 Then
									bFirst1 = False
									lX1# = X1 + sprite1\collisionData[i]\fRealX#
									lY1# = Y1 + sprite1\collisionData[i]\fRealY#
									If bFirst2 Then
										bFirst2 = False
									End If
								Else
									If bFirst2 Then
										bFirst2 = False
									Else
										; Check
										tx1 = sprite1\collisionData[i]\fRealX# + X1
										ty1 = sprite1\collisionData[i]\fRealY# + Y1
										tx3 = sprite2\collisionData[j]\fRealX# + X2
										ty3 = sprite2\collisionData[j]\fRealY# + Y2
										tx2 = lX1
										ty2 = lY1
										tx4 = lX2
										ty4 = lY2
										; If we are not comparing a point
										If (tx1 <> tx2 Or ty1 <> ty2) And (tx3 <> tx4 Or ty3 <> ty4) Then
											dSqr = (ty4-ty3)*(tx2-tx1)-(tx4-tx3)*(ty2-ty1)
											d1 = (ty1-ty3)
											d2 = (tx1-tx3)
											dX = ((tx2-tx1)*d1-(ty2-ty1)*d2)
											dY = ((tx4-tx3)*d1-(ty4-ty3)*d2)
											; Parallel
											If dSqr = 0 Then
												; Coincident
												If dX = 0 And dY = 0 Then
													Return True
												End If
											Else
												d1 = dX / dSqr
												d2 = dY / dSqr
												If (d1 >= 0 And d1 <= 1) And (d2 >= 0 And d2 <= 1) Then
													Return True
												End If
											End If
										End If
									End If
								End If
							; Circle versus circle
							Else If (Not sprite1\collisionData[i]\bPoint) And (Not sprite2\collisionData[j]\bPoint) Then
								dX# = (X1 + sprite1\collisionData[i]\fRealX) - (X2 + sprite2\collisionData[j]\fRealX) 
								dY# = (Y1 + sprite1\collisionData[i]\fRealY) - (Y2 + sprite2\collisionData[j]\fRealY) 
								dSqr# = dX * dX + dY * dY
								If dSqr# < (sprite1\collisionData[i]\fRealRadius + sprite2\collisionData[j]\fRealRadius) * (sprite1\collisionData[i]\fRealRadius + sprite2\collisionData[j]\fRealRadius) Then
									Return True
								End If
							End If
						End If
					End If
					lX2# = X2 + sprite2\collisionData[j]\fRealX#
					lY2# = Y2 + sprite2\collisionData[j]\fRealY#
				End If
			End If
		Next	
		If sprite1\collisionData[i] <> Null Then
			lX1# = X1 + sprite1\collisionData[i]\fRealX#
			lY1# = Y1 + sprite1\collisionData[i]\fRealY#									
		End If
	Next
	Return False
End Function

; ----------------------------------------------------------------------

; Frees the collisions from an image
Function nS_FreeCollisions(Image%)
	img.nS_Image = Object.nS_Image(Image)
	For i = 0 To nS_MAX_COLLISIONS
		If img\collisionData[i] <> Null Then
			Delete img\collisionData[i]
			img\collisionData[i] = Null
		End If
	Next
End Function

; ----------------------------------------------------------------------

; Copies the collisions from an image to another
Function nS_CopyCollisions(Image1%, Image2%)
	nS_FreeCollisions(Image2)
	sprite1.nS_Image = Object.nS_Image(Image1)
	sprite2.nS_Image = Object.nS_Image(Image2)
	For i = 0 To nS_MAX_COLLISIONS
		If sprite1\collisionData[i] <> Null Then
			sprite2\collisionData[i] = New nS_Collision
			sprite2\collisionData[i]\iX = sprite1\collisionData[i]\iX
			sprite2\collisionData[i]\iY = sprite1\collisionData[i]\iY
			sprite2\collisionData[i]\fRadius = sprite1\collisionData[i]\fRadius
			sprite2\collisionData[i]\iType = sprite1\collisionData[i]\iType			
			sprite2\collisionData[i]\bPoint = sprite1\collisionData[i]\bPoint
		End If
	Next	
	nS_UpdateCollisions(Image2)
End Function

; ----------------------------------------------------------------------

; Save collisions from an image to a file
Function nS_SaveCollisions(Image%, sFile$)
	Local img.nS_Image = Object.nS_Image(Image)
	Local Stream = WriteFile(sFile$)
		WriteInt Stream, nS_CurrentVersion
		For i = 0 To nS_MAX_COLLISIONS
			If img\collisionData[i] <> Null Then
				WriteInt Stream, img\collisionData[i]\iX
				WriteInt Stream, img\collisionData[i]\iY
				WriteInt Stream, img\collisionData[i]\iType
				WriteByte Stream, img\collisionData[i]\bPoint
				If Not img\collisionData[i]\bPoint Then
					WriteFloat Stream, img\collisionData[i]\fRadius
				End If		
			End If
		Next
	CloseFile Stream
End Function

; ----------------------------------------------------------------------

; Load collisions from a file to an image
Function nS_LoadCollisions(Image%, sFile$)
	Local img.nS_Image = Object.nS_Image(Image)
	Local header, version, iX, iY, iType, bPoint, fRadius#
	Local Stream = ReadFile(sFile$)
		nS_FreeCollisions(Image)
		i = 0
		version = ReadInt(Stream)
		nS_Assert(version = nS_CurrentVersion, "Incorrect collision data version")
		While Not Eof(Stream)
			iX = ReadInt(Stream)
			iY = ReadInt(Stream)
			iType = ReadInt(Stream)
			bPoint = ReadByte(Stream)
			If Not bPoint Then
				fRadius = ReadFloat(Stream)
				nS_CreateCollision(Image, iX, iY, fRadius, iType)
			Else
				nS_CreateCollision(Image, iX, iY, 0, iType, True)
			End If
		Wend
		nS_UpdateCollisions(Image)
	CloseFile Stream
End Function