; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Globals and Constants
; ======================================================================
Global nS_Millisecs = -1

; ======================================================================
; Different objects
; ======================================================================

; Event object
Type nS_Event
	Field iCategory
	Field bSignaled
End Type

; Timer object
Type nS_Timer
	Field iDelay
	Field iLastMillisecs
	Field bPaused
	Field bEnded
	Field hEndEvent
End Type

; ======================================================================
; Functions
; ======================================================================

; Creates a new event
Function nS_CreateEvent%(iCategory% = -1)
	Local event.nS_Event = New nS_Event
	event\iCategory = iCategory
	Return Handle(event)
End Function

; ----------------------------------------------------------------------

; Signals the event
Function nS_SignalEvent(hEvent%)
	Local event.nS_Event = Object.nS_Event(hEvent)
	event\bSignaled = True
End Function

; ----------------------------------------------------------------------

; Resets the event
Function nS_ResetEvent(hEvent%)
	Local event.nS_Event = Object.nS_Event(hEvent)
	event\bSignaled = False
End Function

; ----------------------------------------------------------------------

; Checks if the event was previously signaled
Function nS_CheckEvent%(hEvent%)
	Local event.nS_Event = Object.nS_Event(hEvent)
	Return event\bSignaled
End Function

; ----------------------------------------------------------------------

; Frees the event from memory
Function nS_FreeEvent(hEvent%)
	Local event.nS_Event = Object.nS_Event(hEvent)
	Delete event
End Function

; ----------------------------------------------------------------------

; Resets all events
Function nS_ResetAllEvents(iCategory% = -1)
	Local event.nS_Event
	For event = Each nS_Event
		If iCategory = -1 Or iCategory = event\iCategory Then
			event\bSignaled = False
		End If
	Next
End Function

; ----------------------------------------------------------------------

; Creates a timer object
Function nS_CreateTimer%(iDelay% = -1, hEndEvent% = -1)
	Local timer.nS_Timer = New nS_Timer
	timer\hEndEvent = hEndEvent
	nS_SetTimer(Handle(timer), iDelay)
	Return Handle(timer)
End Function

; ----------------------------------------------------------------------

; Sets the timer & (re)start it
Function nS_SetTimer(hTimer%, iDelay%, hEndEvent% = -1)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	If nS_Millisecs = -1 Then
		nS_Millisecs = MilliSecs()
	End If
	timer\iLastMillisecs = nS_Millisecs
	timer\iDelay = iDelay
	If hEndEvent <> -1 Then
		timer\hEndEvent = hEndEvent
	End If
	timer\bPaused = False
	timer\bEnded = False
End Function

; ----------------------------------------------------------------------

; Pauses a timer
Function nS_PauseTimer(hTimer%)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	timer\bPaused = True
End Function

; ----------------------------------------------------------------------

; Resumes a timer
Function nS_ResumeTimer(hTimer%)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	timer\bPaused = False
End Function

; ----------------------------------------------------------------------

; Is timer running
Function nS_TimerRunning%(hTimer%)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	Return Not(timer\bPaused Or timer\bEnded)
End Function

; ----------------------------------------------------------------------

; Returns how much time is left
Function nS_TimerTimeLeft%(hTimer%)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	Return timer\iDelay
End Function

; ----------------------------------------------------------------------

; Has the timer ended
Function nS_TimerEnded%(hTimer%)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	Return timer\bEnded
End Function

; ----------------------------------------------------------------------

; Frees the timer object
Function nS_FreeTimer(hTimer%)
	Local timer.nS_Timer = Object.nS_Timer(hTimer)
	Delete timer
End Function

; ----------------------------------------------------------------------

; Updates every timer
Function nS_UpdateTimers()
	Local timer.nS_Timer
	Local diff
	nS_Millisecs = MilliSecs()
	For timer = Each nS_Timer
		; If the timer hasn't reached end
		If (Not timer\bEnded) Then
			; If the timer isn't paused
			If (Not timer\bPaused) Then
				diff = nS_Millisecs - timer\iLastMillisecs
				timer\iDelay = timer\iDelay - diff
				If timer\iDelay <= 0 Then
					timer\bEnded = True
					; Signal End Event if present
					If timer\hEndEvent <> -1 Then
						nS_SignalEvent(timer\hEndEvent)
					End If
				End If
			End If
			timer\iLastMillisecs = nS_Millisecs
		End If
	Next
End Function