; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Globals and Constants
; ======================================================================

; Types
Const nS_TYPE_BUTTON = 1
Const nS_TYPE_TEXTBOX = 2
Const nS_TYPE_PANEL = 3
Const nS_TYPE_CHECKBOX = 4

; Panels
Const nS_PANEL_PLAIN = 1
Const nS_PANEL_RAISED = 2
Const nS_PANEL_SUNKEN = 3
Const nS_PANEL_SLIM_RAISED = 4
Const nS_PANEL_SLIM_SUNKEN = 5

; Styles
Const nS_STYLE_WINDOWS = 1
Const nS_STYLE_NS2GUI = 2

; GUI Images
Const nS_GUIIREZ = 512
Const nS_MAXGUIIMAGE = 32
Global nS_GUIImageCount = -1
Global nS_GUIImage[nS_MAXGUIIMAGE]
Global nS_GUIIX[nS_MAXGUIIMAGE]
Global nS_GUIIY[nS_MAXGUIIMAGE]
Global nS_GUIIH[nS_MAXGUIIMAGE]

; Focus
Global nS_BUTTONFOCUS = False
Global nS_TabCount = 0
Global nS_TabFocus = -1
Global nS_WasTabbed = False

; Mouse events
Global nS_MouseClicked = False
Global nS_MouseDown = False
Global nS_MouseX
Global nS_MouseY

; Fonts
Global nS_fntWBlack
Global nS_fntWGray
Global nS_fntWWhite
Global nS_fntWToolTip
Global nS_fntNS2GUI
Global nS_fntNS2GUIRed

; Timing
Global nS_DeltaTime
Global nS_LastTime

; ======================================================================
; Different objects
; ======================================================================

; Control object
Type nS_Control
		; Image & Style
		Field hImage%
		Field iStyle%
		; Positioning and size
		Field fTop#
		Field fLeft#
		Field fWidth#
		Field fHeight#
		Field bEnabled%
		; Mouse position on it
		Field iMX%
		Field iMY%
		; Focus
		Field bHasFocus%
		Field iTabIndex%
		; Text
		Field sCaption$
		Field bCaption%
		Field sToolTip$
		Field bToolTip%
		Field iTimeToolTip%
		; Events (possible)
		Field bClick%
		Field bGetFocus%
		Field bType%
		; Events (happened)
		Field bClicked%
		Field bGotFocus%
		Field bLostFocus%
		Field iKeyHit%
		; Control type specific
		Field iControlType%
		Field hControlData%
End Type

; Button extended data
Type nS_ButtonData
	Field oControl.nS_Control
	Field hImgDown%
	Field bDefault%
	Field iDownTimer%
	Field sprIcon%
End Type

; Textbox extended data
Type nS_TextBoxData
	Field oControl.nS_Control
	Field sText$
	Field iCursor%
	Field bCursor%
	Field iCursorTimer%
	Field hImgDisabled%
	Field iMaxLength%
End Type

; Checkbox extended data
Type nS_CheckBoxData
	Field oControl.nS_Control
	Field hImgUncDis%
	Field hImgChkEnb%
	Field hImgChkDis%
	Field bChecked%
End Type

; Image data
Type nS_ImageData
	Field buffer%
	Field himg%
	Field offx%
	Field offy%
End Type

; ======================================================================
; Functions
; ======================================================================

; Generic Control Methods
; ----------------------------------------------------------------------

; Set the position of a control
Function nS_ControlSetPosition(hControl%, fLeft#, fTop#)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	oControl\fTop = fTop
	oControl\fLeft = fLeft
End Function

; Set the caption of a control
Function nS_ControlSetCaption(hControl%, sCaption$)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	nS_Assert(oControl\bCaption, "Control type cannot have caption")
	oControl\sCaption = sCaption
	; Calculate Checkbox width
	If oControl\iControlType = nS_TYPE_CHECKBOX Then
		oControl\fWidth = 27 + nS_StringWidth(oControl\sCaption, nS_fntWBlack)
	End If
End Function

; Set the tooltip text of a control
Function nS_ControlSetTooltip(hControl%, sToolTip$)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	nS_Assert(oControl\bToolTip, "Control type cannot have tooltip")
	oControl\sToolTip = sToolTip
End Function

; Enable and disable a control
Function nS_ControlEnable(hControl%, bEnable%)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	oControl\bEnabled = bEnable
End Function

; Check if a control has been clicked
Function nS_ControlClicked%(hControl%)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	Local ret%
	nS_Assert(oControl\bClick, "Control cannot be clicked")
	ret = oControl\bClicked
	oControl\bClicked = False
	Return ret
End Function

; Check if a control got focus
Function nS_ControlGotFocus%(hControl%)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	Local ret%
	nS_Assert(oControl\bGetFocus, "Control cannot receive/lose focus")
	ret = oControl\bGotFocus
	oControl\bGotFocus = False
	Return ret
End Function

; Check if a control lost focus
Function nS_ControlLostFocus%(hControl%)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	Local ret%
	nS_Assert(oControl\bGetFocus, "Control cannot receive/lose focus")
	ret = oControl\bLostFocus
	oControl\bLostFocus = False
	Return ret
End Function

; Check if a control has been typed on
Function nS_ControlKeyHit%(hControl%)
	Local oControl.nS_Control = Object.nS_Control(hControl)
	Local ret%
	nS_Assert(oControl\bType, "Control cannot be typed in")
	ret = oControl\iKeyHit
	oControl\iKeyHit = 0
	Return ret
End Function

; Clear events of a specified control or all controls
Function nS_ControlClearEvents(hControl% = -1)
	Local oControl.nS_Control
	If hControl = -1 Then
		; Clear all events of all controls
		For oControl = Each nS_Control
			oControl\iKeyHit = 0
			oControl\bLostFocus = False
			oControl\bGotFocus = False
			oControl\bClicked = False
		Next
	Else
		; Clear all events for specified control
		oControl = Object.nS_Control(hControl)
		oControl\iKeyHit = 0
		oControl\bLostFocus = False
		oControl\bGotFocus = False
		oControl\bClicked = False
	End If
End Function

; Buttons Methods
; ----------------------------------------------------------------------

; Create a new button
Function nS_CreateButton%(fWidth#, fHeight#, iStyle% = nS_STYLE_WINDOWS)
	Local oControl.nS_Control = New nS_Control
	Local oButton.nS_ButtonData = New nS_ButtonData
	Local i.nS_ImageData
	Local x, y, c
	oButton\oControl = oControl
	; Generic data
	oControl\fWidth = fWidth
	oControl\fHeight = fHeight
	oControl\bEnabled = True
	; Type and style
	oControl\iControlType = nS_TYPE_BUTTON
	oControl\iStyle = iStyle
	; Data possible
	oControl\bToolTip = True
	oControl\bCaption = True
	; Tab
	oControl\iTabIndex = nS_TabCount
	nS_TabCount = nS_TabCount + 1
	; Events possible
	oControl\bClick = True
	oControl\bGetFocus = True
	; Button data
	oControl\hControlData = Handle(oButton)
	; Create button images
	Select iStyle
		; Windows style buttons
		Case nS_STYLE_WINDOWS
			nS_LoadWindowsFonts()
			; Button Up (base image)
			i = nS_AllocateImage(fWidth, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			nS_RectImage(1, 1, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(1, fHeight - 2, fWidth - 2, 1, i, nS_RGBA(128, 128, 128)) 
			nS_RectImage(fWidth - 2, 1, 1, fHeight - 3, i, nS_RGBA(128, 128, 128)) 
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(64, 64, 64)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(64, 64, 64)) 
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
			; Button down
			i = nS_AllocateImage(fWidth, fHeight)
			oButton\hImgDown = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oButton\hImgDown)
			nS_RectImage(2, 2, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200))
			WritePixelFast 1 + i\offx, fHeight - 2 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			WritePixelFast fWidth - 2 + i\offx, 1 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(64, 64, 64)) 
			nS_RectImage(0, 1, 1, fHeight - 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(128, 128, 128)) 
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(128, 128, 128)) 
			UnlockBuffer nS_ImageBuffer(oButton\hImgDown)
		; nSprite 2 style buttons
		Case nS_STYLE_NS2GUI
			nS_LoadNS2GUIFonts()
			; Button Up (base image)
			i = nS_AllocateImage(fWidth, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			nS_Gradient(0, 0, fWidth, fHeight, i)
			nS_Bevel(1, 1, fWidth-2, fHeight-2, i, 2, 155, 175, 255, 0.25, 155, 175, 255, -0.25)
			For x = 0 To fWidth - 1
				y = 0
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
				y = fHeight - 1
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
			Next
			For y = 0 To fHeight - 1
				x = 0
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
				x = fWidth - 1
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
			Next
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
			; Button down
			i = nS_AllocateImage(fWidth, fHeight)
			oButton\hImgDown = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oButton\hImgDown)
			nS_Gradient(0, 0, fWidth, fHeight, i)
			nS_Bevel(1, 1, fWidth-2, fHeight-2, i, 2, 155, 175, 255, -0.25, 155, 175, 255, 0.25)
			For x = 0 To fWidth - 1
				y = 0
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
				y = fHeight - 1
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
			Next
			For y = 0 To fHeight - 1
				x = 0
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
				x = fWidth - 1
				c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
				nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) - 79, nS_RGBG(c) - 88, nS_RGBB(c) - 128, i\Buffer)
			Next
			UnlockBuffer nS_ImageBuffer(oButton\hImgDown)
	End Select
	; Return
	Return Handle(oControl)
End Function

; Set the default button
Function nS_ButtonSetDefault(hButton%)
	Local oControl.nS_Control = Object.nS_Control(hButton)
	Local oButton.nS_ButtonData
	For oButton = Each nS_ButtonData
		oButton\bDefault = False
	Next
	nS_Assert(oControl\iControlType = nS_TYPE_BUTTON, "Control is not a Button")
	oButton = Object.nS_ButtonData(oControl\hControlData)
	oButton\bDefault = True
End Function

; Set an icon for a button
Function nS_ButtonSetIcon(hButton%, sprIcon%)
	Local oControl.nS_Control = Object.nS_Control(hButton)
	Local oButton.nS_ButtonData = Object.nS_ButtonData(oControl\hControlData)
	Local iHeight% = nS_ImageHeight(sprIcon)
	Local iWidth% = nS_ImageWidth(sprIcon)
	nS_HandleImage sprIcon, 0, iHeight/2
	oButton\sprIcon = sprIcon
	If iHeight > oControl\fHeight - 12 Then
		iWidth = (Float(oControl\fHeight - 12)/Float(iHeight))*Float(iWidth)
		iHeight = oControl\fHeight - 12
	End If
	If iWidth > oControl\fWidth - 12 Then
		iHeight = (Float(oControl\fWidth - 12)/Float(iWidth))*Float(iHeight)
		iWidth = oControl\fWidth - 12
	End If
	nS_ResizeImage sprIcon, iWidth, iHeight
End Function

; Textboxes Methods
; ----------------------------------------------------------------------

; Create a new text box
Function nS_CreateTextBox%(fWidth#, iStyle% = nS_STYLE_WINDOWS)
	Local oControl.nS_Control = New nS_Control
	Local oTextbox.nS_TextBoxData = New nS_TextBoxData
	Local i.nS_ImageData
	Local fHeight#
	Local x, y, c
	oTextbox\oControl = oControl
	; Generic data
	oControl\fWidth = fWidth
	oControl\bEnabled = True
	; Type and style
	oControl\iControlType = nS_TYPE_TEXTBOX
	oControl\iStyle = iStyle
	; Data possible
	oControl\bToolTip = True
	oControl\bCaption = False
	; Tab
	oControl\iTabIndex = nS_TabCount
	nS_TabCount = nS_TabCount + 1
	; Events possible
	oControl\bClick = True
	oControl\bGetFocus = True
	oControl\bType = True
	; Textbox data
	oControl\hControlData = Handle(oTextbox)
	; Create textbox images
	Select iStyle
		; Windows style textbox
		Case nS_STYLE_WINDOWS
			; Fixed height
			fHeight = 20
			oControl\fHeight = fHeight
			nS_LoadWindowsFonts()
			; Textbox image
			i = nS_AllocateImage(fWidth, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			nS_RectImage(2, 2, fWidth - 4, fHeight - 4, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(1, fHeight - 2, fWidth - 2, 1, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(fWidth - 2, 1, 1, fHeight - 3, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
			; Disabled text box
			i = nS_AllocateImage(fWidth, fHeight)
			oTextBox\hImgDisabled = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oTextBox\hImgDisabled)
			nS_RectImage(2, 2, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200))
			WritePixelFast 1 + i\offx, fHeight - 2 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			WritePixelFast fWidth - 2 + i\offx, 1 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))
			UnlockBuffer nS_ImageBuffer(oTextBox\hImgDisabled)
		; nSprite 2 style textbox
		Case nS_STYLE_NS2GUI
			; Fixed height
			fHeight = 28
			oControl\fHeight = fHeight
			nS_LoadWindowsFonts()
			; Textbox image
			i = nS_AllocateImage(fWidth, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			nS_Gradient(0, 0, fWidth, fHeight, i)
			nS_Bevel(0, 0, fWidth, fHeight, i, 3, 155, 175, 255, -0.5, 155, 175, 255, 0.5)
			For x = 3 To fWidth - 4
				For y = 3 To fHeight - 4
					c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
					nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) + 20, nS_RGBG(c) + 25, nS_RGBB(c) + 40, i\Buffer)
				Next
			Next
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
			; Disabled text box
			oTextBox\hImgDisabled = oControl\hImage
	End Select
	; Return
	Return Handle(oControl)
End Function

; Set the textbox' text
Function nS_TextBoxSetText(hTextBox%, sText$)
	Local oControl.nS_Control = Object.nS_Control(hTextBox)
	Local oTextbox.nS_TextBoxData = Object.nS_TextBoxData(oControl\hControlData)
	oTextBox\sText = sText
End Function

; Set the cursor position
Function nS_TextBoxSetCursor(hTextBox%, iCursor%)
	Local oControl.nS_Control = Object.nS_Control(hTextBox)
	Local oTextbox.nS_TextBoxData = Object.nS_TextBoxData(oControl\hControlData)
	oTextBox\iCursor = iCursor
End Function

; Get the text
Function nS_TextBoxGetText$(hTextBox%)
	Local oControl.nS_Control = Object.nS_Control(hTextBox)
	Local oTextbox.nS_TextBoxData = Object.nS_TextBoxData(oControl\hControlData)
	Return oTextBox\sText
End Function

; Set the maximum length of input
Function nS_TextBoxSetMaxLen(hTextBox%, iMaxLength%)
	Local oControl.nS_Control = Object.nS_Control(hTextBox)
	Local oTextbox.nS_TextBoxData = Object.nS_TextBoxData(oControl\hControlData)
	oTextBox\iMaxLength = iMaxLength
End Function

; Panel Methods
; ----------------------------------------------------------------------

; Create a panel
Function nS_CreatePanel(fWidth#, fHeight#, iPanel% = nS_PANEL_PLAIN, iStyle% = nS_STYLE_WINDOWS)
	Local oControl.nS_Control = New nS_Control
	Local oTextbox.nS_TextBoxData = New nS_TextBoxData
	Local i.nS_ImageData
	oTextbox\oControl = oControl
	; Generic data
	oControl\fWidth = fWidth
	oControl\fHeight = fHeight
	oControl\bEnabled = False
	; Type and style
	oControl\iControlType = nS_TYPE_PANEL
	oControl\iStyle = iStyle
	; Data possible
	oControl\bToolTip = False
	oControl\bCaption = False
	; Tab
	oControl\iTabIndex = nS_TabCount
	nS_TabCount = nS_TabCount + 1
	; Events possible
	oControl\bClick = False
	oControl\bGetFocus = False
	oControl\bType = False
	; Create image
	Select iStyle
		; Windows style
		Case nS_STYLE_WINDOWS
			i = nS_AllocateImage(fWidth, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			Select iPanel
				Case nS_PANEL_PLAIN
					nS_RectImage(0, 0, fWidth, fHeight, i, nS_RGBA(212, 208, 200)) 
				Case nS_PANEL_RAISED
					nS_RectImage(1, 1, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200)) 
					nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(255, 255, 255)) 
					nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(255, 255, 255)) 
					nS_RectImage(1, fHeight - 2, fWidth - 2, 1, i, nS_RGBA(128, 128, 128)) 
					nS_RectImage(fWidth - 2, 1, 1, fHeight - 3, i, nS_RGBA(128, 128, 128)) 
					nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(64, 64, 64)) 
					nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(64, 64, 64)) 
				Case nS_PANEL_SUNKEN
					nS_RectImage(2, 2, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200))
					WritePixelFast 1 + i\offx, fHeight - 2 + i\offy, nS_RGBA(212, 208, 200), i\buffer
					WritePixelFast fWidth - 2 + i\offx, 1 + i\offy, nS_RGBA(212, 208, 200), i\buffer
					nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
					nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
					nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
					nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
					nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
					nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))			
				Case nS_PANEL_SLIM_RAISED
					nS_RectImage(1, 1, fWidth - 2, fHeight - 2, i, nS_RGBA(212, 208, 200))
					nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(255, 255, 255)) 
					nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(255, 255, 255)) 
					nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(128, 128, 128))
					nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(128, 128, 128))
				Case nS_PANEL_SLIM_SUNKEN
					nS_RectImage(1, 1, fWidth - 2, fHeight - 2, i, nS_RGBA(212, 208, 200))
					nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128)) 
					nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128)) 
					nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255))
					nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))				
			End Select
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
		; nSprite 2 style
		Case nS_STYLE_NS2GUI
			i = nS_AllocateImage(fWidth, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			Select iPanel
				Case nS_PANEL_PLAIN
					nS_Gradient(0, 0, fWidth, fHeight, i)
				Case nS_PANEL_RAISED
					nS_Gradient(0, 0, fWidth, fHeight, i)
					nS_Bevel(0, 0, fWidth, fHeight, i, 2, 155, 175, 255, 0.25, 155, 175, 255, -0.25)
				Case nS_PANEL_SUNKEN
					nS_Gradient(0, 0, fWidth, fHeight, i)
					nS_Bevel(0, 0, fWidth, fHeight, i, 2, 155, 175, 255, -0.25, 155, 175, 255, 0.25)
				Case nS_PANEL_SLIM_RAISED
					nS_Gradient(0, 0, fWidth, fHeight, i)
					nS_Bevel(0, 0, fWidth, fHeight, i, 1, 155, 175, 255, 0.25, 155, 175, 255, -0.25)
				Case nS_PANEL_SLIM_SUNKEN
					nS_Gradient(0, 0, fWidth, fHeight, i)
					nS_Bevel(0, 0, fWidth, fHeight, i, 1, 155, 175, 255, -0.25, 155, 175, 255, 0.25)
			End Select
			UnlockBuffer nS_ImageBuffer(oControl\hImage)			
	End Select
	; Return
	Return Handle(oControl)
End Function

; CheckBox Methods
; ----------------------------------------------------------------------

; Create a new checkbox
Function nS_CreateCheckBox%(sCaption$ = "", iStyle% = nS_STYLE_WINDOWS)
	Local oControl.nS_Control = New nS_Control
	Local oCheckbox.nS_CheckBoxData = New nS_CheckBoxData
	Local i.nS_ImageData
	Local fWidth#, fHeight#
	oCheckbox\oControl = oControl
	oControl\sCaption = sCaption
	; Generic data
	oControl\fWidth = 0
	oControl\fHeight = 0
	oControl\bEnabled = True
	; Type and style
	oControl\iControlType = nS_TYPE_CHECKBOX
	oControl\iStyle = iStyle
	; Data possible
	oControl\bToolTip = True
	oControl\bCaption = True
	; Tab
	oControl\iTabIndex = nS_TabCount
	nS_TabCount = nS_TabCount + 1
	; Events possible
	oControl\bClick = True
	oControl\bGetFocus = True
	oControl\bType = False
	; Checkbox data
	oControl\hControlData = Handle(oCheckbox)
	; Create checkbox images
	Select iStyle
		; Windows style checkbox
		Case nS_STYLE_WINDOWS
			nS_LoadWindowsFonts()
			; Fixed size
			fWidth = 13
			fHeight = fWidth
			; Enabled unchecked checkbox image
			i = nS_AllocateImage(fWidth*2, fHeight*2)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			nS_RectImage(2, 2, fWidth - 4, fHeight - 4, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(1, fHeight - 2, fWidth - 2, 1, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(fWidth - 2, 1, 1, fHeight - 3, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
			; Disabled unchecked checkbox image
			i\offx = i\offx + fWidth
			oCheckBox\hImgUncDis = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oCheckBox\hImgUncDis)
			nS_RectImage(2, 2, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200))
			WritePixelFast 1 + i\offx, fHeight - 2 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			WritePixelFast fWidth - 2 + i\offx, 1 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))
			UnlockBuffer nS_ImageBuffer(oCheckBox\hImgUncDis)
			; Enabled checked checkbox image
			i\offy = i\offy + fHeight
			oCheckBox\hImgChkEnb = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oCheckBox\hImgChkEnb)
			nS_RectImage(2, 2, fWidth - 4, fHeight - 4, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(1, fHeight - 2, fWidth - 2, 1, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(fWidth - 2, 1, 1, fHeight - 3, i, nS_RGBA(212, 208, 200)) 
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))
			; check
			nS_RectImage(3, 5, 1, 3, i, nS_RGBA(0, 0, 0))
			nS_RectImage(4, 6, 1, 3, i, nS_RGBA(0, 0, 0))
			nS_RectImage(5, 7, 1, 3, i, nS_RGBA(0, 0, 0))
			nS_RectImage(6, 6, 1, 3, i, nS_RGBA(0, 0, 0))
			nS_RectImage(7, 5, 1, 3, i, nS_RGBA(0, 0, 0))
			nS_RectImage(8, 4, 1, 3, i, nS_RGBA(0, 0, 0))
			nS_RectImage(9, 3, 1, 3, i, nS_RGBA(0, 0, 0))
			UnlockBuffer nS_ImageBuffer(oCheckBox\hImgChkEnb)			
			; Disabled checked checkbox image
			i\offx = i\offx - fWidth
			oCheckBox\hImgChkDis = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oCheckBox\hImgChkDis)
			nS_RectImage(2, 2, fWidth - 3, fHeight - 3, i, nS_RGBA(212, 208, 200))
			WritePixelFast 1 + i\offx, fHeight - 2 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			WritePixelFast fWidth - 2 + i\offx, 1 + i\offy, nS_RGBA(212, 208, 200), i\buffer
			nS_RectImage(0, fHeight - 1, fWidth, 1, i, nS_RGBA(255, 255, 255)) 
			nS_RectImage(fWidth - 1, 0, 1, fHeight - 1, i, nS_RGBA(255, 255, 255))
			nS_RectImage(0, 0, fWidth - 1, 1, i, nS_RGBA(128, 128, 128))
			nS_RectImage(0, 1, 1, fHeight - 2, i, nS_RGBA(128, 128, 128))
			nS_RectImage(1, 1, fWidth - 3, 1, i, nS_RGBA(64, 64, 64))
			nS_RectImage(1, 2, 1, fHeight - 4, i, nS_RGBA(64, 64, 64))
			; check
			nS_RectImage(3, 5, 1, 3, i, nS_RGBA(128, 128, 128))
			nS_RectImage(4, 6, 1, 3, i, nS_RGBA(128, 128, 128))
			nS_RectImage(5, 7, 1, 3, i, nS_RGBA(128, 128, 128))
			nS_RectImage(6, 6, 1, 3, i, nS_RGBA(128, 128, 128))
			nS_RectImage(7, 5, 1, 3, i, nS_RGBA(128, 128, 128))
			nS_RectImage(8, 4, 1, 3, i, nS_RGBA(128, 128, 128))
			nS_RectImage(9, 3, 1, 3, i, nS_RGBA(128, 128, 128))			
			UnlockBuffer nS_ImageBuffer(oCheckBox\hImgChkDis)
			; Get height and width
			oControl\fHeight = fHeight
			oControl\fWidth = fWidth + 14 + nS_StringWidth(oControl\sCaption, nS_fntWBlack)
		; nSprite 2 style checkbox
		Case nS_STYLE_NS2GUI
			; Fixed height & height
			fHeight = 15
			fWidth = fHeight
			nS_LoadNS2GUIFonts()
			; Empty Checkbox image
			i = nS_AllocateImage(fWidth*2, fHeight)
			oControl\hImage = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oControl\hImage)
			nS_Gradient(0, 0, fWidth, fHeight, i)
			nS_Bevel(0, 0, fWidth, fHeight, i, 2, 155, 175, 255, -0.5, 155, 175, 255, 0.5)
			For x = 3 To fWidth - 4
				For y = 3 To fHeight - 4
					c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
					nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) + 20, nS_RGBG(c) + 25, nS_RGBB(c) + 40, i\Buffer)
				Next
			Next
			UnlockBuffer nS_ImageBuffer(oControl\hImage)
			; Checked checkbox image
			i\offx = i\offx + fWidth
			oCheckBox\hImgChkEnb = nS_CreateSubImage(i\himg, i\offx, i\offy, fWidth, fHeight)
			LockBuffer nS_ImageBuffer(oCheckBox\hImgChkEnb)
			nS_Gradient(0, 0, fWidth, fHeight, i)
			nS_Bevel(0, 0, fWidth, fHeight, i, 2, 155, 175, 255, -0.5, 155, 175, 255, 0.5)
			For x = 3 To fWidth - 4
				For y = 3 To fHeight - 4
					c = ReadPixelFast(i\offx+x, i\offy+y, i\buffer)
					nS_WriteCapPixel(i\offx+x, i\offy+y, nS_RGBR(c) + 20, nS_RGBG(c) + 25, nS_RGBB(c) + 40, i\Buffer)
				Next
			Next
			nS_RectImage(3, 3, 1, 2, i, nS_RGBA(30, 80, 10))
			nS_RectImage(4, 3, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(5, 4, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(6, 5, 1, 5, i, nS_RGBA(30, 80, 10))
			nS_RectImage(7, 6, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(8, 5, 1, 5, i, nS_RGBA(30, 80, 10))
			nS_RectImage(9, 4, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(10, 3, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(11, 3, 1, 2, i, nS_RGBA(30, 80, 10))
			nS_RectImage(3, 10, 1, 2, i, nS_RGBA(30, 80, 10))
			nS_RectImage(4, 9, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(5, 8, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(9, 8, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(10, 9, 1, 3, i, nS_RGBA(30, 80, 10))
			nS_RectImage(11, 10, 1, 2, i, nS_RGBA(30, 80, 10))
			UnlockBuffer nS_ImageBuffer(oCheckBox\hImgChkEnb)	
			; Get height and width
			oControl\fHeight = fHeight
			oControl\fWidth = fWidth + 15 + nS_StringWidth(oControl\sCaption, nS_fntNS2GUI)
	End Select
	; Return
	Return Handle(oControl)
End Function

; Set the checkbox check state
Function nS_CheckBoxSetChecked(hCheckBox%, bChecked%)
	Local oControl.nS_Control = Object.nS_Control(hCheckBox)
	Local oCheckbox.nS_CheckBoxData = Object.nS_CheckBoxData(oControl\hControlData)
	oCheckBox\bChecked = bChecked
End Function

; Check if the checkbox is checked
Function nS_CheckBoxIsChecked%(hCheckBox%)
	Local oControl.nS_Control = Object.nS_Control(hCheckBox)
	Local oCheckbox.nS_CheckBoxData = Object.nS_CheckBoxData(oControl\hControlData)
	Return oCheckBox\bChecked
End Function

; Label Methods
; ----------------------------------------------------------------------

; Draw a simple label
Function nS_DrawLabel(fLeft#, fTop#, sText$, bCenterX% = False, bCenterY% = False, iStyle% = nS_STYLE_WINDOWS, bDisabled = False)
	Select iStyle
		Case nS_STYLE_WINDOWS
			nS_LoadWindowsFonts()
			If bDisabled Then
				nS_Text fLeft+1, fTop+1, sText, bCenterX, bCenterY, nS_fntWWhite
				nS_Text fLeft, fTop, sText, bCenterX, bCenterY, nS_fntWGray
			Else
				nS_Text fLeft, fTop, sText, bCenterX, bCenterY, nS_fntWBlack
			End If
			; Post-render font order
			nS_Text -10, 0, " ", False, False, nS_fntWToolTip
		Case nS_STYLE_NS2GUI
			nS_LoadNS2GUIFonts()
			If bDisabled Then
				nS_Text fLeft, fTop, sText, bCenterX, bCenterY, nS_fntNS2GUIRed
			Else
				nS_Text fLeft, fTop, sText, bCenterX, bCenterY, nS_fntNS2GUI
			End If
	End Select
End Function

; Global Methods
; ----------------------------------------------------------------------

; Update all controls and events
Function nS_UpdateGUI()
	Local oControl.nS_Control
	Local oOthers.nS_Control
	Local oButton.nS_ButtonData
	Local oTextBox.nS_TextBoxData
	Local oCheckBox.nS_CheckBoxData
	Local bTab%, bEnter%, bMouseOver%, bSpace%, iKey%
	nS_BUTTONFOCUS = False
	bTab = KeyHit(15)
	bEnter = KeyHit(28) Or KeyHit(156)
	bSpace = KeyHit(57)
	iKey = GetKey()
	; Tab
	If nS_TabFocus = -1 Then
		If bTab Then
			nS_WasTabbed = True
			bTab = False
			nS_TabFocus = nS_TabFocus + 1
		End If
	End If
	; Timing
	nS_DeltaTime = MilliSecs() - nS_LastTime
	nS_LastTime = MilliSecs()
	; Mouse
	nS_MouseClicked = MouseHit(1)
	nS_MouseDown = MouseDown(1)
	nS_MouseX = MouseX()
	nS_MouseY = MouseY()
	For oControl = Each nS_Control
		bMouseOver = False
		; Mouse Over & Tool tip
		If nS_MouseX >= oControl\fLeft And nS_MouseX < oControl\fLeft + oControl\fWidth And nS_MouseY >= oControl\fTop And nS_MouseY < oControl\fTop + oControl\fHeight Then
			If oControl\bToolTip Then
				If oControl\iMX = nS_MouseX And oControl\iMY = nS_MouseY Then
					If oControl\iTimeToolTip > 0 Then
						oControl\iTimeToolTip = oControl\iTimeToolTip - nS_DeltaTime
					End If
				Else
					If oControl\iTimeToolTip > 0 Then
						oControl\iMX = nS_MouseX
						oControl\iMY = nS_MouseY
						oControl\iTimeToolTip = 1000
					End If
				End If
			End If
			bMouseOver = True
		Else
			oControl\iTimeToolTip = 1000
		End If
		; Clicked
		If oControl\bClick Then
			If nS_MouseClicked Then
				If bMouseOver Then
					If oControl\bEnabled And oControl\bGetFocus Then
						nS_TabFocus = oControl\iTabIndex
						oControl\bClicked = True
						oControl\iTimeToolTip = 1000
					End If
				End If
			End If
		End If
		; Tab indexes
		If oControl\iTabIndex = nS_TabFocus Then
			If (Not oControl\bHasFocus) Then
				If oControl\bGetFocus And oControl\bEnabled Then
					oControl\bGotFocus = True
					oControl\bHasFocus = True
					For oOthers = Each nS_Control
						If Not(oOthers = oControl) Then
							If oOthers\bHasFocus Then
								oOthers\bHasFocus = False
								oOthers\bLostFocus = True
							End If
						End If
					Next
				Else
					nS_TabFocus = nS_TabFocus + 1
					If nS_TabFocus = nS_TabCount Then
						nS_TabFocus = 0
					End If
				End If
			End If
			; Tab
			If bTab Then
				nS_WasTabbed = True
				bTab = False
				oControl\bHasFocus = False
				oControl\bLostFocus = True
				nS_TabFocus = nS_TabFocus + 1
				If nS_TabFocus = nS_TabCount Then
					nS_TabFocus = 0
				End If		
			End If
		End If
		; Key typed
		If oControl\bType And oControl\bHasFocus Then
			oControl\iKeyHit = iKey
		End If
		; Button specifics
		If oControl\iControlType = nS_TYPE_BUTTON Then
			oButton = Object.nS_ButtonData(oControl\hControlData)
			; Button focus
			If oControl\bHasFocus And (Not oButton\bDefault) Then
				nS_BUTTONFOCUS = True
			End If
			; Down timer update
			If oButton\iDownTimer > 0 Then
				oButton\iDownTimer = oButton\iDownTimer - nS_DeltaTime
			End If
		End If
		; Textbox specifics
		If oControl\iControlType = nS_TYPE_TEXTBOX Then
			oTextBox = Object.nS_TextBoxData(oControl\hControlData)
			; Click focus, but no events
			If oControl\bClicked Then
				oControl\bClicked = False
			End If
			; Focus'ed
			If oControl\bHasFocus Then
				; Cursor update
				If oTextBox\iCursorTimer <= 0 Then
					oTextBox\bCursor = Not oTextBox\bCursor
					oTextBox\iCursorTimer = 550
				Else
					oTextBox\iCursorTimer = oTextBox\iCursorTimer - nS_DeltaTime
				End If
				; Typing
				; Plain and extended characters
				If ((iKey >= 32 And iKey <= 126) Or (iKey >= 128 And iKey <= 254)) And ((Len(oTextBox\sText) < oTextBox\iMaxLength) Or (oTextBox\iMaxLength = 0)) Then
					oTextBox\sText = Left(oTextBox\sText, oTextBox\iCursor) + Chr(iKey) + Right(oTextBox\sText, Len(oTextBox\sText) - oTextBox\iCursor)
					oTextBox\iCursor = oTextBox\iCursor + 1
					oControl\iTimeToolTip = 1000
				; Backspace
				ElseIf iKey = 8 Then
					If oTextBox\iCursor > 0 Then
						oTextBox\sText = Left(oTextBox\sText, oTextBox\iCursor - 1) + Right(oTextBox\sText, Len(oTextBox\sText) - oTextBox\iCursor)
						oTextBox\iCursor = oTextBox\iCursor - 1
						oControl\iTimeToolTip = 1000
					End If
				; Delete
				ElseIf iKey = 4 Then
					If oTextBox\iCursor < Len(oTextBox\sText) Then
						oTextBox\sText = Left(oTextBox\sText, oTextBox\iCursor) + Right(oTextBox\sText, Len(oTextBox\sText) - oTextBox\iCursor - 1)
						oControl\iTimeToolTip = 1000
					End If
				; Left Arrow
				ElseIf iKey = 31 Then
					If oTextBox\iCursor > 0 Then
						oTextBox\iCursor = oTextBox\iCursor - 1
						oControl\iTimeToolTip = 1000
					End If
				; Right Arrow
				ElseIf iKey = 30 Then
					If oTextBox\iCursor < Len(oTextBox\sText) Then
						oTextBox\iCursor = oTextBox\iCursor + 1
						oControl\iTimeToolTip = 1000
					End If
				; Home
				ElseIf iKey = 1 Then
					oTextBox\iCursor = 0
					oControl\iTimeToolTip = 1000
				; End
				ElseIf iKey = 2 Then
					oTextBox\iCursor = Len(oTextBox\sText)
					oControl\iTimeToolTip = 1000
				End If
			End If
		End If
		; Checkbox specifics
		If oControl\iControlType = nS_TYPE_CHECKBOX Then
			oCheckBox = Object.nS_CheckBoxData(oControl\hControlData)
			; When on focus
			If oControl\bHasFocus Then
				; Button focus to prevent default
				nS_BUTTONFOCUS = True
				; Clicked
				If oControl\bClicked And nS_MouseClicked And bMouseOver Then
					oCheckBox\bChecked = Not oCheckBox\bChecked
				End If
				; Switch when enter or space
				If bEnter Or bSpace Then
					oCheckBox\bChecked = Not oCheckBox\bChecked
					oControl\bClicked = True
				End If
			End If
			
		End If
	Next
	; Button post-update
	For oButton = Each nS_ButtonData
		; Button enter or space
		If bEnter Then
			If oButton\oControl\bHasFocus Or (oButton\bDefault And (Not nS_BUTTONFOCUS) And oButton\oControl\bEnabled) Then
				oButton\oControl\bClicked = True
				oButton\iDownTimer = 100
			End If
		End If
		If bSpace Then
			If oButton\oControl\bHasFocus Then
				oButton\oControl\bClicked = True
				oButton\iDownTimer = 100
			End If
		End If
	Next
End Function

; Draw all controls
Function nS_DrawGUI()
	Local oControl.nS_Control
	Local oButton.nS_ButtonData
	Local oTextBox.nS_TextBoxData
	Local oCheckBox.nS_CheckBoxData
	Local oFocus.nS_Control
	Local oToolTip.nS_Control
	Local sx%, sy%, x%, y%, ex%, ey%, iDown%, xOff%
	Local caption$
	; Render
	For oControl = Each nS_Control
		iDown = 0
		xOff = 0
		; Focus
		If oControl\bHasFocus Then
			oFocus = oControl
		End If
		; Tool tip
		If oControl\iTimeToolTip <= 0 And oControl\bToolTip And (Not oControl\sToolTip = "") Then
			oToolTip = oControl
		End If		
		Select oControl\iControlType
			; Button
			Case nS_TYPE_BUTTON
				oButton = Object.nS_ButtonData(oControl\hControlData)
				; Draw button image
				If (oControl\bHasFocus And (nS_MouseDown And nS_MouseX >= oControl\fLeft And nS_MouseX < oControl\fLeft + oControl\fWidth And nS_MouseY >= oControl\fTop And nS_MouseY < oControl\fTop + oControl\fHeight)) Or oButton\iDownTimer > 0 Then
					iDown = 1
					nS_DrawImage oButton\hImgDown, oControl\fLeft, oControl\fTop
				Else
					nS_DrawImage oControl\hImage, oControl\fLeft, oControl\fTop
				End If
				; Style features
				Select oControl\iStyle
					; Windows style features
					Case nS_STYLE_WINDOWS
						; If it has icon
						If oButton\sprIcon > 0 Then
							xOff = nS_ImageWidth(oButton\sprIcon)+8
						End If
						; Text
						If Not oControl\sCaption = "" Then
							sx = 1
							ex = Len(oControl\sCaption) + 1
							sy = False
							ey = oControl\fWidth - xOff - 6
							While nS_StringWidth(Trim(Mid(oControl\sCaption, sx, ex - sx)), nS_fntWBlack) > ey
								If sy Then
									sx = sx + 1
								Else
									ex = ex - 1
								End If
								sy = Not sy
							Wend
							caption = Trim(Mid(oControl\sCaption, sx, ex - sx))
							sx = oControl\fLeft + xOff + Floor((oControl\fWidth-xOff)/2) + iDown - 1
							sy = oControl\fTop + Floor(oControl\fHeight/2) + iDown - 2
							If oControl\bEnabled Then
								nS_Text sx, sy, caption, True, True, nS_fntWBlack
							Else
								nS_Text sx + 1, sy + 1, caption, True, True, nS_fntWWhite
								nS_Text sx, sy, caption, True, True, nS_fntWGray
							End If
						End If
						; Focus or default rect
						If oControl\bHasFocus Or (oButton\bDefault And (Not nS_BUTTONFOCUS) And oControl\bEnabled) Then
							nS_Color 0, 0, 0
							nS_Rect oControl\fLeft, oControl\fTop, oControl\fWidth + 1, oControl\fHeight + 1, False, 1
						End If
					; nSprite 2 style features
					Case nS_STYLE_NS2GUI
						; If it has icon
						If oButton\sprIcon > 0 Then
							xOff = nS_ImageWidth(oButton\sprIcon)+8
						End If
						; Text
						If Not oControl\sCaption = "" Then
							sx = 1
							ex = Len(oControl\sCaption) + 1
							sy = False
							ey = oControl\fWidth - xOff - 8
							While nS_StringWidth(Trim(Mid(oControl\sCaption, sx, ex - sx)), nS_fntNS2GUI) > ey
								If sy Then
									sx = sx + 1
								Else
									ex = ex - 1
								End If
								sy = Not sy
							Wend
							caption = Trim(Mid(oControl\sCaption, sx, ex - sx))
							sx = oControl\fLeft + xOff + Floor((oControl\fWidth-xOff)/2) + iDown
							sy = oControl\fTop + Floor(oControl\fHeight/2) + iDown - 2
							If oControl\bEnabled Then
								nS_Text sx, sy, caption, True, True, nS_fntNS2GUI
							Else
								nS_Text sx, sy, caption, True, True, nS_fntNS2GUIRed
							End If
						End If
						; Focus or default rect
						If oControl\bHasFocus Or (oButton\bDefault And (Not nS_BUTTONFOCUS) And oControl\bEnabled) Then
							nS_Color 255, 255, 255, 0.6
							nS_Rect oControl\fLeft, oControl\fTop, oControl\fWidth + 1, oControl\fHeight + 1, False, 1
						End If										
				End Select
			; Textbox
			Case nS_TYPE_TEXTBOX
				oTextBox = Object.nS_TextBoxData(oControl\hControlData)
				; Enabled
				If oControl\bEnabled Then
					nS_DrawImage oControl\hImage, oControl\fLeft, oControl\fTop
				; Disabled
				Else
					nS_DrawImage oTextBox\hImgDisabled, oControl\fLeft, oControl\fTop				
				End If
				; Style features
				Select oControl\iStyle
					; Windows style features
					Case nS_STYLE_WINDOWS
						; Enabled
						If oControl\bEnabled Then
							sx = 1
							If oTextBox\sText <> "" Then
								; Longer than textbox
								If nS_StringWidth(oTextBox\sText, nS_fntWBlack) <= oControl\fWidth - 8 Then
									nS_Text oControl\fLeft + 4, oControl\fTop + 3, oTextBox\sText, False, False, nS_fntWBlack
								Else
									; Cursor is within the textbox or doesn't have focus
									If (nS_StringWidth(Left(oTextBox\sText, oTextBox\iCursor), nS_fntWBlack) <= oControl\fWidth - 12) Or (Not oControl\bHasFocus) Then
										sx = Len(oTextBox\sText)
										While nS_StringWidth(Left(oTextBox\sText, sx), nS_fntWBlack) > oControl\fWidth - 12
											sx = sx - 1
										Wend
										nS_Text oControl\fLeft + 4, oControl\fTop + 3, Left(oTextBox\sText, sx), False, False, nS_fntWBlack
										sx = 1
									Else
									; Cursor is out of the textbox and has focus
										sx = 1
										While nS_StringWidth(Mid(oTextBox\sText, sx, oTextBox\iCursor - sx + 1), nS_fntWBlack) > oControl\fWidth - 12
											sx = sx + 1
										Wend
										ex = oTextBox\iCursor
										While (nS_StringWidth(Mid(oTextBox\sText, sx, ex - sx + 1), nS_fntWBlack) <= oControl\fWidth - 12) And (ex < Len(oTextBox\sText))
											ex = ex + 1
										Wend
										nS_Text oControl\fLeft + 4, oControl\fTop + 3, Mid(oTextBox\sText, sx, ex - sx + 1), False, False, nS_fntWBlack
									End If
								End If
							End If
							; Cursor
							If oControl\bHasFocus And oTextBox\bCursor Then
								nS_Color 0, 0, 0
								sx = oControl\fLeft + 5 + nS_StringWidth(Mid(oTextBox\sText, sx, oTextBox\iCursor - sx + 1), nS_fntWBlack)
								nS_Line sx, oControl\fTop + 3, sx, oControl\fTop + oControl\fHeight - 3
							End If
						; Disabled
						Else
							If oTextBox\sText <> "" Then
								; Longer than textbox
								If nS_StringWidth(oTextBox\sText, nS_fntWBlack) <= oControl\fWidth - 12 Then
									nS_Text oControl\fLeft + 4, oControl\fTop + 4, oTextBox\sText, False, False, nS_fntWGray
								Else
									sx = Len(oTextBox\sText)
									While nS_StringWidth(Left(oTextBox\sText, sx), nS_fntWBlack) > oControl\fWidth - 12
										sx = sx - 1
									Wend
									nS_Text oControl\fLeft + 4, oControl\fTop + 3, Left(oTextBox\sText, sx), False, False, nS_fntWGray
								End If
							End If
						End If
					; nSprite 2 style features
					Case nS_STYLE_NS2GUI
						; Enabled
						If oControl\bEnabled Then
							sx = 1
							If oTextBox\sText <> "" Then
								; Longer than textbox
								If nS_StringWidth(oTextBox\sText, nS_fntNS2GUI) <= oControl\fWidth - 8 Then
									nS_Text oControl\fLeft + 4, oControl\fTop + 4, oTextBox\sText, False, False, nS_fntNS2GUI
								Else
									; Cursor is within the textbox or doesn't have focus
									If (nS_StringWidth(Left(oTextBox\sText, oTextBox\iCursor), nS_fntNS2GUI) <= oControl\fWidth - 12) Or (Not oControl\bHasFocus) Then
										sx = Len(oTextBox\sText)
										While nS_StringWidth(Left(oTextBox\sText, sx), nS_fntNS2GUI) > oControl\fWidth - 12
											sx = sx - 1
										Wend
										nS_Text oControl\fLeft + 4, oControl\fTop + 4, Left(oTextBox\sText, sx), False, False, nS_fntNS2GUI
										sx = 1
									Else
									; Cursor is out of the textbox and has focus
										sx = 1
										While nS_StringWidth(Mid(oTextBox\sText, sx, oTextBox\iCursor - sx + 1), nS_fntNS2GUI) > oControl\fWidth - 12
											sx = sx + 1
										Wend
										ex = oTextBox\iCursor
										While (nS_StringWidth(Mid(oTextBox\sText, sx, ex - sx + 1), nS_fntNS2GUI) <= oControl\fWidth - 12) And (ex < Len(oTextBox\sText))
											ex = ex + 1
										Wend
										nS_Text oControl\fLeft + 4, oControl\fTop + 4, Mid(oTextBox\sText, sx, ex - sx + 1), False, False, nS_fntNS2GUI
									End If
								End If
							End If
							; Cursor
							If oControl\bHasFocus And oTextBox\bCursor Then
								nS_Color 0, 0, 0, 0.5
								sx = oControl\fLeft + 3 + nS_StringWidth(Mid(oTextBox\sText, sx, oTextBox\iCursor - sx + 1), nS_fntNS2GUI)
								nS_Rect sx, oControl\fTop + 4, 3, oControl\fHeight - 9
							End If
						; Disabled
						Else
							If oTextBox\sText <> "" Then
								; Longer than textbox
								If nS_StringWidth(oTextBox\sText, nS_fntNS2GUI) <= oControl\fWidth - 12 Then
									nS_Text oControl\fLeft + 4, oControl\fTop + 4, oTextBox\sText, False, False, nS_fntNS2GUIRed
								Else
									sx = Len(oTextBox\sText)
									While nS_StringWidth(Left(oTextBox\sText, sx), nS_fntNS2GUI) > oControl\fWidth - 12
										sx = sx - 1
									Wend
									nS_Text oControl\fLeft + 4, oControl\fTop + 4, Left(oTextBox\sText, sx), False, False, nS_fntNS2GUIRed
								End If
							End If
						End If
				End Select
			; Panel
			Case nS_TYPE_PANEL
				nS_DrawImage oControl\hImage, oControl\fLeft, oControl\fTop			
			; Checkbox
			Case nS_TYPE_CHECKBOX
				oCheckBox = Object.nS_CheckBoxData(oControl\hControlData)
				; Style features
				Select oControl\iStyle
					; Windows style features
					Case nS_STYLE_WINDOWS	
						sx = oControl\fLeft + 20
						If oControl\bEnabled Then
							If oCheckBox\bChecked Then
								nS_DrawImage oCheckBox\hImgChkEnb, oControl\fLeft, oControl\fTop
							Else
								nS_DrawImage oControl\hImage, oControl\fLeft, oControl\fTop
							End If
							nS_Text sx, oControl\fTop, oControl\sCaption, False, False, nS_fntWBlack
						Else
							If oCheckBox\bChecked Then
								nS_DrawImage oCheckBox\hImgChkDis, oControl\fLeft, oControl\fTop
							Else
								nS_DrawImage oCheckBox\hImgUncDis, oControl\fLeft, oControl\fTop
							End If
							nS_Text sx+1, oControl\fTop+1, oControl\sCaption, False, False, nS_fntWWhite
							nS_Text sx, oControl\fTop, oControl\sCaption, False, False, nS_fntWGray
						End If
					; Windows style features
					Case nS_STYLE_NS2GUI	
						sx = oControl\fLeft + 20
						If oControl\bEnabled Then
							If oCheckBox\bChecked Then
								nS_DrawImage oCheckBox\hImgChkEnb, oControl\fLeft, oControl\fTop
							Else
								nS_DrawImage oControl\hImage, oControl\fLeft, oControl\fTop
							End If
							nS_Text sx, oControl\fTop - 3, oControl\sCaption, False, False, nS_fntNS2GUI
						Else
							If oCheckBox\bChecked Then
								nS_DrawImage oCheckBox\hImgChkEnb, oControl\fLeft, oControl\fTop
							Else
								nS_DrawImage oControl\hImage, oControl\fLeft, oControl\fTop
							End If
							nS_Text sx, oControl\fTop - 3, oControl\sCaption, False, False, nS_fntNS2GUIRed
						End If
				End Select
		End Select
	Next
	; Post-render focus features
	If (Not(oFocus = Null)) And (nS_WasTabbed) Then
		Select oFocus\iStyle
			; Windows style dashes
			Case nS_STYLE_WINDOWS
				; On a button
				If oFocus\iControlType = nS_TYPE_BUTTON Then
					nS_Color 0, 0, 0
					sx = oFocus\fLeft + 3
					sy = oFocus\fTop + 3
					ex = oFocus\fLeft + oFocus\fWidth - 5
					ey = oFocus\fTop + oFocus\fHeight - 5
					If (ex - sx) Mod 2 = 1 Then
						ex = ex + 1
					End If
					If (ey - sy) Mod 2 = 1 Then
						ey = ey + 1
					End If
					For x = sx+1 To ex Step 2
						nS_Plot x, sy
						nS_Plot x, ey
					Next
					For y = sy+1 To ey Step 2
						nS_Plot sx, y
						nS_Plot ex, y
					Next
				End If
				; On a checkbox
				If oFocus\iControlType = nS_TYPE_CHECKBOX Then
					nS_Color 0, 0, 0
					sx = oFocus\fLeft + 17
					sy = oFocus\fTop + 1
					ex = oFocus\fLeft + oFocus\fWidth - 1
					ey = oFocus\fTop + oFocus\fHeight + 1
					For x = sx+1 To ex Step 2
						nS_Plot x, sy
						nS_Plot x, ey
					Next
					For y = sy+1 To ey Step 2
						nS_Plot sx, y
						nS_Plot ex, y
					Next
				End If
			; nSprite 2 feature
			Case nS_STYLE_NS2GUI
				; On a button
				If oFocus\iControlType = nS_TYPE_BUTTON Then
					nS_Color 255, 255, 255, 0.2
					nS_Rect oFocus\fLeft, oFocus\fTop, oFocus\fWidth, oFocus\fHeight, True
				End If
				; On a checkbox
				If oFocus\iControlType = nS_TYPE_CHECKBOX Then
					nS_Color 255, 255, 255, 0.2
					nS_Rect oFocus\fLeft - 1, oFocus\fTop - 1, oFocus\fWidth, 17, True
				End If
			End Select
	End If
	; Post-render icons
	For oControl = Each nS_Control
		iDown = 0
		Select oControl\iControlType
			; Button
			Case nS_TYPE_BUTTON
				oButton = Object.nS_ButtonData(oControl\hControlData)
				; Down
				If (oControl\bHasFocus And (nS_MouseDown And nS_MouseX >= oControl\fLeft And nS_MouseX < oControl\fLeft + oControl\fWidth And nS_MouseY >= oControl\fTop And nS_MouseY < oControl\fTop + oControl\fHeight)) Or oButton\iDownTimer > 0 Then
					iDown = 1
				End If
				; Style features
				Select oControl\iStyle
					; Windows style features
					Case nS_STYLE_WINDOWS
						; If it has icon
						If oButton\sprIcon > 0 Then
							; If Enabled or Disabled
							If oControl\bEnabled Then
								nS_ImageAlpha oButton\sprIcon, 1
							Else
								nS_ImageAlpha oButton\sprIcon, 0.5
							End If
							; Left-aligned if there's a caption
							If Not oControl\sCaption = "" Then
								nS_DrawImage oButton\sprIcon, oControl\fLeft + iDown + 5, oControl\fTop + Floor(oControl\fHeight/2) + iDown
							; Center-aligned if there's no caption
							Else
								nS_DrawImage oButton\sprIcon, oControl\fLeft + iDown + oControl\fWidth/2 - nS_ImageWidth(oButton\sprIcon)/2, oControl\fTop + Floor(oControl\fHeight/2) + iDown
							End If
						End If
					; nSprite 2 style features
					Case nS_STYLE_NS2GUI
						; If it has icon
						If oButton\sprIcon > 0 Then
							; If Enabled or Disabled
							If oControl\bEnabled Then
								nS_ImageColor oButton\sprIcon, 255, 255, 255
							Else
								nS_ImageColor oButton\sprIcon, 255, 0, 0
							End If
							; Left-aligned if there's a caption
							If Not oControl\sCaption = "" Then
								nS_DrawImage oButton\sprIcon, oControl\fLeft + iDown + 6, oControl\fTop + Floor(oControl\fHeight/2) + iDown
							; Center-aligned if there's no caption
							Else
								nS_DrawImage oButton\sprIcon, oControl\fLeft + iDown + oControl\fWidth/2 - nS_ImageWidth(oButton\sprIcon)/2, oControl\fTop + Floor(oControl\fHeight/2) + iDown
							End If
						End If
				End Select
		End Select
	Next
	; Post-render tool tip
	If oToolTip <> Null Then
		Select oToolTip\iStyle
			; Windows style tool tip
			Case nS_STYLE_WINDOWS
				x = nS_StringWidth(oToolTip\sToolTip, nS_fntWToolTip) + 6
				y = nS_StringHeight(oToolTip\sToolTip, nS_fntWToolTip) + 6
				nS_Color 255, 255, 225
				nS_Rect oToolTip\iMX, oToolTip\iMY + 22, x-1, y-1, True
				nS_Color 0, 0, 0
				nS_Rect oToolTip\iMX, oToolTip\iMY + 22, x, y, False, 1
				nS_Text oToolTip\iMX + 2, oToolTip\iMY + 22, oToolTip\sToolTip, False, False, nS_fntWToolTip
			; nSprite 2 style tool tip
			Case nS_STYLE_NS2GUI
				x = nS_StringWidth(oToolTip\sToolTip, nS_fntNS2GUI) + 10
				y = nS_StringHeight(oToolTip\sToolTip, nS_fntNS2GUI) + 6
				nS_Color 0, 0, 0, 0.7
				nS_Rect oToolTip\iMX - 4, oToolTip\iMY + 23, x, y, True
				nS_Text oToolTip\iMX + 1, oToolTip\iMY + 24, oToolTip\sToolTip, False, False, nS_fntNS2GUI
		End Select
	End If
	; Post-render order
	nS_Plot -10, 0
	If nS_fntWBlack Then
		nS_Text -10, 0, " ", False, False, nS_fntWBlack
		nS_Text -10, 0, " ", False, False, nS_fntWWhite
		nS_Text -10, 0, " ", False, False, nS_fntWGray
		nS_Text -10, 0, " ", False, False, nS_fntWToolTip
	End If
	If nS_fntNS2GUI Then
		nS_Text -10, 0, " ", False, False, nS_fntNS2GUI
		nS_Text -10, 0, " ", False, False, nS_fntNS2GUIRed
	End If
End Function

; ============================================
; SYSTEM FUNCTIONS
; ============================================

Function nS_LoadWindowsFonts()
	If nS_fntWBlack = 0 Then
		nS_fntWBlack = nS_LoadFont("nS_Windows.fnt")
		nS_fntWGray = nS_LoadFont("nS_Windows.fnt")
		nS_fntWWhite = nS_LoadFont("nS_Windows.fnt")
		nS_fntWToolTip = nS_LoadFont("nS_WindowsBig.fnt")
		nS_FontColor nS_fntWBlack, 0, 0, 0
		nS_FontColor nS_fntWGray, 128, 128, 128
		nS_FontColor nS_fntWWhite, 255, 255, 255
		nS_FontColor nS_fntWToolTip, 0, 0, 0
	End If
End Function

Function nS_LoadNS2GUIFonts()
	If Not nS_fntNS2GUI Then
		nS_fntNS2GUI = nS_LoadFont("nS_NS2GUI.fnt")
		nS_fntNS2GUIRed = nS_LoadFont("nS_NS2GUI.fnt")
		nS_FontColor nS_fntNS2GUI, 90, 120, 190
		nS_FontColor nS_fntNS2GUIRed, 180, 40, 55
	End If
End Function

Function nS_AllocateImage.nS_ImageData(iWidth%, iHeight%)
	Local img.nS_ImageData = New nS_ImageData
	Local bCreateNew = False
	If nS_GUIImageCount = -1 Then
		bCreateNew = True
	Else
		If nS_GUIIX[nS_GUIImageCount] + iWidth > nS_GUIIREZ Then
			If nS_GUIIY[nS_GUIImageCount] + nS_GUIIH[nS_GUIImageCount] + iHeight > nS_GUIIREZ Then
				bCreateNew = True
			End If
		End If
	End If
	If bCreateNew Then
		nS_GUIImageCount = nS_GUIImageCount + 1
		newRez = nS_GUIIREZ;
		; Exception for images larger than rez
		While iHeight > newRez or iWidth > newRez
			newRez = newRez * 2
		Wend
		nS_GUIImage[nS_GUIImageCount] = nS_CreateImage(newRez, newRez, 1, 1+2)
		nS_GUIIX[nS_GUIImageCount] = iWidth
		nS_GUIIY[nS_GUIImageCount] = 0
		nS_GUIIH[nS_GUIImageCount] = iHeight
		img\offx = 0
		img\offy = 0
	Else
		If nS_GUIIX[nS_GUIImageCount] + iWidth <= nS_GUIIREZ Then
			If iHeight > nS_GUIIH[nS_GUIImageCount] Then
				nS_GUIIH[nS_GUIImageCount] = iHeight
			End If
		Else
			nS_GUIIX[nS_GUIImageCount] = 0
			nS_GUIIY[nS_GUIImageCount] = nS_GUIIY[nS_GUIImageCount] + nS_GUIIH[nS_GUIImageCount]
			nS_GUIIH[nS_GUIImageCount] = iHeight
		End If
		img\offx = nS_GUIIX[nS_GUIImageCount]
		img\offy = nS_GUIIY[nS_GUIImageCount]
		nS_GUIIX[nS_GUIImageCount] = nS_GUIIX[nS_GUIImageCount] + iWidth
	End If
	img\buffer = nS_ImageBuffer(nS_GUIImage[nS_GUIImageCount])
	img\himg = nS_GUIImage[nS_GUIImageCount]
	Return img
End Function

Function nS_RGBA%(R, G, B, A = 255)
	Return ((A Shl 24) Or (R Shl 16) Or (G Shl 8) Or (B))
End Function

Function nS_RectImage(x, y, w, h, i.nS_ImageData, argb)
	Local ix, iy, tx, ty
	x = x + i\offx
	y = y + i\offy
	tx = x + w - 1
	ty = y + h - 1
	For ix = x To tx
		For iy = y To ty
			WritePixelFast ix, iy, argb, i\buffer
		Next
	Next
End Function

Function nS_Gradient(x, y, w, h, i.nS_ImageData)
	Local sr# = 30, sg# = 130, sb# = 200
	Local cr#, cg#, cb#
	Local dr# = 85, dg# = 55, db# = 35
	Local ix, iy
	Local d#, md#, s#
	s = Float(w)/Float(h)
	md = w*w+s*h*h
	For ix = 0 To w
		For iy = 0 To h
			d = ix*ix+s*iy*iy
			d = 1-d/md
			cr = sr + dr*d
			cg = sg + dg*d
			cb = sb + db*d
			WritePixelFast i\offx + x + ix, i\offy + y + iy, nS_RGBA(cr, cg, cb), i\buffer
		Next
	Next
End Function

Function nS_Bevel(x, y, w, h, i.nS_ImageData, size, tlr#, tlg#, tlb#, tla#, brr#, brg#, brb#, bra#)
	Local c, ix, iy, m
	; Top
	For xi = 0 To w - size - 1
		For yi = 0 To size - 1
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + tlr*tla, nS_RGBG(c) + tlg*tla, nS_RGBB(c) + tlb*tla, i\Buffer)
		Next
	Next
	; Left
	For xi = 0 To size - 1
		For yi = size To h - size - 1
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + tlr*tla, nS_RGBG(c) + tlg*tla, nS_RGBB(c) + tlb*tla, i\Buffer)
		Next
	Next
	; Bottom
	For xi = size To w - 1
		For yi = h - size To h - 1
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + brr*bra, nS_RGBG(c) + brg*bra, nS_RGBB(c) + brb*bra, i\Buffer)
		Next
	Next
	; Right
	For xi = w - size To w - 1
		For yi = size To h - size - 1
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + brr*bra, nS_RGBG(c) + brg*bra, nS_RGBB(c) + brb*bra, i\Buffer)
		Next
	Next
	; Top right corner
	m = 0
	For yi = 0 To size - 1
		For xi = w - size To w - 1 - m
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + tlr*tla, nS_RGBG(c) + tlg*tla, nS_RGBB(c) + tlb*tla, i\Buffer)
		Next
		For xi = w - m To w - 1
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + brr*bra, nS_RGBG(c) + brg*bra, nS_RGBB(c) + brb*bra, i\Buffer)
		Next
		m = m + 1
	Next
	; Bottom left corner
	m = 0
	For yi = h - size To h - 1
		For xi = 0 To size - m - 2
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + tlr*tla, nS_RGBG(c) + tlg*tla, nS_RGBB(c) + tlb*tla, i\Buffer)
		Next
		For xi = size - 1 - m To size - 1
			c = ReadPixelFast(i\offx+xi+x, i\offy+yi+y, i\buffer)
			nS_WriteCapPixel(i\offx+xi+x, i\offy+yi+y, nS_RGBR(c) + brr*bra, nS_RGBG(c) + brg*bra, nS_RGBB(c) + brb*bra, i\Buffer)
		Next
		m = m + 1
	Next
End Function

Function nS_WriteCapPixel(x, y, r, g, b, buffer)
	If r > 255 Then
		r = 255
	End If
	If g > 255 Then
		g = 255
	End If
	If b > 255 Then
		b = 255
	End If
	If r < 0 Then
		r = 0
	End If
	If g < 0 Then
		g = 0
	End If
	If b < 0 Then
		b = 0
	End If	
	WritePixelFast x, y, nS_RGBA(r, g, b), buffer
End Function