; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

Type nS_Image
	; Mesh
	Field mesh
	Field surface
	; Data
	Field width
	Field height
	Field name$
	Field flags
	Field frames
	; Color
	Field r
	Field g
	Field b
	Field a#
	; Blending
	Field fx
	Field blend
	; Mask
	Field maskRed
	Field maskGreen
	Field maskBlue
	; Position
	Field x#
	Field y#
	Field centered
	; Animation
	Field mode
	Field speed#
	Field startf#
	Field endf#
	Field frame#
	; Vertex color
	Field vertexcolored[4]
	Field vertexr[4]
	Field vertexg[4]
	Field vertexb[4]
	Field vertexa#[4]
	; Vertex position
	Field vertox#[4]
	Field vertoy#[4]
	; Scale & Rotation
	Field scalex#
	Field scaley#
	Field angle#
	; Base Image
	Field image
	; Texture
	Field texture
	Field size
	Field u1#
	Field v1#
	Field u2#
	Field v2#
	; Instances
	Field instanceMax
	Field instanceID[nS_MaxInstances]
	Field instanceUsed[nS_MaxInstances]
	; Collision Data
	Field collisionData.nS_Collision[nS_MAX_COLLISIONS]
	Field fColXMin#
	Field fColYMin#
	Field fColXMax#
	Field fColYMax#
	; Lighting
	Field bLightAffected
	Field bUseVertexLighting
End Type

; Creates a new image from scratch
Function nS_CreateImage%(width%, height%, frames% = 1, flags% = 5)
	Local image.nS_Image = New nS_Image
	Local i, x, y, buffer

	image\width = width
	image\height = height
	
	image\flags = flags
	image\frames = frames

	image\size = nS_GetTextureSize(width, height)
	image\u2 = Float(width) / Float(image\size)
	image\v2 = Float(height) / Float(image\size)
	image\texture = CreateTexture(width, height, flags, frames)
	
	nS_DefaultImageConstructor(image)

	Return Handle(image)
End Function

; Loads an image from a file
Function nS_LoadImage%(bmpfile$, flags% = 5, bFastLoad% = True)
	Return nS_LoadAnimImage(bmpfile, -1, -1, 0, 1, flags, bFastLoad)
End Function

; Loads an animated image frmo a file
Function nS_LoadAnimImage%(bmpfile$, width%, height%, frst%, count%, flags% = 5, bFastLoad% = True)
	Local image.nS_Image = New nS_Image
	Local x, y, i, c

	; Base image
	If Not bFastLoad Then
		; Single frame image
		If width = -1 Then
			image\image = LoadImage(bmpfile)
			width = ImageWidth(image\image)
			height = ImageHeight(image\image)
		; Animated image
		Else
			image\image = LoadAnimImage(bmpfile, width, height, frst, count)
		End If
	End If

	; Fast loading	
	If bFastLoad Then
		; Single frame image
		If width = -1 Then
			image\texture = LoadTexture(bmpfile, flags)
			image\width = TextureWidth(image\texture)
			image\height = TextureHeight(image\texture)
		; Animated image
		Else
			image\texture = LoadAnimTexture(bmpfile, flags, width, height, frst, count)
			image\width = width
			image\height = height
			image\size = width
		End If
		image\size = image\width
	; Non-Fast loading						
	Else
		image\size = nS_GetTextureSize(width, height)
		; Base image is loaded and the size is incorrect, copy for cleaner result
		image\texture = CreateTexture(image\size, image\size, flags, count)
		For i = frst To frst + count - 1
			LockBuffer ImageBuffer(image\image, i)
			LockBuffer TextureBuffer(image\texture, i)
				For x = 0 To image\size - 1
					For y = 0 To image\size - 1
						If x < width And y < height Then
							c = ReadPixelFast(x, y, ImageBuffer(image\image, i))
							If c = -16777216 Then
								c = 0
							End If
							WritePixelFast x, y, c, TextureBuffer(image\texture, i)
						Else
							WritePixelFast x, y, 0, TextureBuffer(image\texture, i)
						End If
					Next
				Next
			UnlockBuffer ImageBuffer(image\image, i)
			UnlockBuffer TextureBuffer(image\texture, i)
		Next
		image\width = image\size
		image\height = image\size
	End If

	image\u2 = 1
	image\v2 = 1
	image\flags = flags
	image\frames = count
	
	image\name = bmpfile

	nS_DefaultImageConstructor(image)
	
	Return Handle(image)
End Function

; Copies an image
Function nS_CopyImage%(oldimage%)
	Local image.nS_Image = Object.nS_Image(oldimage)
	Local newimage.nS_Image = New nS_Image
	Local i, x, y, c
	
	newimage\texture = CreateTexture(image\width, image\height, image\flags, image\frames)
	For i = 0 To image\frames - 1
		LockBuffer TextureBuffer(image\texture, i)
		LockBuffer TextureBuffer(newimage\texture, i)
			For x = 0 To image\width - 1
				For y = 0 To image\height - 1
					c = ReadPixelFast(x, y, TextureBuffer(image\texture, i))
					If c = -16777216 Then
						c = 0
					End If
					WritePixelFast x, y, c, TextureBuffer(newimage\texture, i)
				Next
			Next
		UnlockBuffer TextureBuffer(image\texture, i)
		UnlockBuffer TextureBuffer(newimage\texture, i)	
	Next
	
	If image\image Then
		newimage\image = CopyImage(image\image)
	End If
	
	nS_DefaultImageConstructor(newimage)

	nS_CopyImageData(image, newimage)
	
	nS_ImageBlend(Handle(newimage), newimage\blend)
	nS_ImageAlpha(Handle(newimage), newimage\a)
	nS_ImageFX(Handle(newimage), newimage\fx)
	
	Return Handle(newimage)
End Function

; Saves the image to disk
Function nS_SaveImage(image%, bmpfile$, frame% = 0)
	Local img.nS_Image = Object.nS_Image(image)
	Local bmp = CreateImage(img\width, img\height)
	LockBuffer TextureBuffer(img\texture, frame)
	LockBuffer ImageBuffer(bmp)
		For x = 0 To img\width - 1
			For y = 0 To img\height - 1
				c = ReadPixelFast(x, y, TextureBuffer(img\texture, frame))
				If c = -16777216 Then
					c = 0
				End If
				WritePixelFast x, y, c, ImageBuffer(bmp)
			Next
		Next
	UnlockBuffer TextureBuffer(img\texture, frame)
	UnlockBuffer ImageBuffer(bmp)
	SaveImage bmp, bmpfile
	FreeImage bmp
End Function

; Grabs an image
Function nS_GrabImage(image%, ox%, oy%, frame% = 0)
	Local img.nS_Image = Object.nS_Image(image)
	Local bmp
	Local bframe = frame
	Local px, py
	If img\image Then
		bmp = img\image
		GrabImage bmp, ox, oy, frame
	End If
	LockBuffer TextureBuffer(img\texture, frame)
	LockBuffer BackBuffer()
		For x = 0 To img\width - 1
			For y = 0 To img\height - 1
				px = x+ox-img\x
				py = y+oy-img\y
				If px < 0 Or py < 0 Or px >= GraphicsWidth() Or py >= GraphicsHeight() Then
					c = 0
				Else
					c = ReadPixelFast(px, py, BackBuffer())
				End If
				If c = -16777216 Then
					c = 0
				End If
				WritePixelFast x, y, c, TextureBuffer(img\texture, frame)
			Next
		Next
	UnlockBuffer TextureBuffer(img\texture, frame)
	UnlockBuffer BackBuffer()
End Function

; Returns the image's buffer
Function nS_ImageBuffer%(image%, frame% = 0)
	Local img.nS_Image = Object.nS_Image(image)
	Return TextureBuffer(img\texture, frame)
End Function

; "Draw" the image, position it for the camera and set up instance
Function nS_DrawImage(image%, x#, y#, frame% = -1)
	Local img.nS_Image = Object.nS_Image(image)
	Local i = 0, id = -1
	Local fx#, fy#
	Local ix#[4], iy#[4], fAngle#, iTemp#, fPosX#[4], fPosY#[4]
	While id = -1
		If Not img\instanceUsed[i] Then
			id = i
		End If
		i = i + 1
		If i > nS_MaxInstances Then
			nS_Assert(False, "Maximum number of instances drawn for image (" + nS_MaxInstances + ")")
			Return
		End If
	Wend
	; Calculate coordinates
	If img\centered Then
		fx# = -0.5 - Float(img\width) * img\scalex / 2.0 + nS_OriginX
		fy# = -0.5 - Float(img\height) * img\scaley / 2.0 + nS_OriginY
	Else
		fx# = -0.5 - Float(img\x) * img\scalex + nS_OriginX
		fy# = -0.5 - Float(img\y) * img\scaley + nS_OriginY
	End If
	; Instance
	img\instanceUsed[id] = True
	If id > img\instanceMax Then
		img\instanceMax = id
	End If
	; The instance isn't there
	If Not (img\instanceID[id] > 0) Then
		img\instanceID[id] = AddVertex(img\surface, 0, 0, 0, img\u1, img\v1)
		AddVertex(img\surface, 0, 0, 0, img\u1, img\v2)
		AddVertex(img\surface, 0, 0, 0, img\u2, img\v2)
		AddVertex(img\surface, 0, 0, 0, img\u2, img\v1)
		AddTriangle(img\surface, img\instanceID[id], img\instanceID[id] + 3, img\instanceID[id] + 1)
		AddTriangle(img\surface, img\instanceID[id] + 2, img\instanceID[id] + 1, img\instanceID[id] + 3)
	End If
	ix[1] = fx + img\vertox[nS_VERTEX_TOPLEFT]
	iy[1] = fy + img\vertoy[nS_VERTEX_TOPLEFT]
	ix[2] = fx + img\vertox[nS_VERTEX_BOTTOMLEFT]
	iy[2] = fy + Float(img\height) * img\scaley + img\vertoy[nS_VERTEX_BOTTOMLEFT]
	ix[3] = fx + Float(img\width) * img\scalex + img\vertox[nS_VERTEX_BOTTOMRIGHT]
	iy[3] = fy + Float(img\height) * img\scaley + img\vertoy[nS_VERTEX_BOTTOMRIGHT]
	ix[4] = fx + Float(img\width) * img\scalex + img\vertox[nS_VERTEX_TOPRIGHT]
	iy[4] = fy + img\vertoy[nS_VERTEX_TOPRIGHT]
	fAngle = img\angle
	For i = 1 To 4
		If fAngle = 0 Then
			fx = ix[i]
			fy = iy[i]
		Else
			fx = Cos(fAngle) * ix[i] - Sin(fAngle) * iy[i]
			fy = Sin(fAngle) * ix[i] + Cos(fAngle) * iy[i]
		End If
		ix[i] = fx
		iy[i] = fy	
		VertexCoords img\surface, img\instanceID[id]+i-1, ix[i]+x+nS_OriginX, iy[i]+y+nS_OriginY, nS_ZBuffer
		If img\vertexcolored[i] Then
			VertexColor img\surface, img\instanceID[id]+i-1, img\vertexr[i], img\vertexg[i], img\vertexb[i], img\vertexa[i]
		End If
	Next
	; Set texture coords
	VertexTexCoords img\surface, img\instanceID[id], img\u1, img\v1
	VertexTexCoords img\surface, img\instanceID[id]+1, img\u1, img\v2
	VertexTexCoords img\surface, img\instanceID[id]+2, img\u2, img\v2
	VertexTexCoords img\surface, img\instanceID[id]+3, img\u2, img\v1
	; Texture from frame or animation
	If frame = -1 Then
		Select img\mode
			Case nS_ANIM_STOP
			Case nS_ANIM_LOOP
				img\frame = img\frame + img\speed * nS_AnimationTween#
				If img\speed > 0 Then
					If img\frame >= img\endf + 1 Then
						img\frame = img\startf 
					End If
				Else
					If img\frame < img\endf Then
						img\frame = img\startf + 0.99999
					End If
				End If	
			Case nS_ANIM_PINGPONG
				img\frame = img\frame + img\speed * nS_AnimationTween#
				If img\speed > 0 Then
					If img\frame >= img\endf + 1 Then
						img\frame = img\endf - 0.00001
						iTemp = img\startf
						img\startf = img\endf
						img\endf = iTemp
						img\speed = img\speed * -1
						img\mode = nS_ANIM_ONESHOT
					End If
				Else
					If img\frame <= img\endf Then
						img\frame = img\endf + 0.00001
						iTemp = img\startf
						img\startf = img\endf
						img\endf = iTemp
						img\speed = img\speed * -1
						img\mode = nS_ANIM_ONESHOT
					End If
				End If	
			Case nS_ANIM_ONESHOT
				img\frame = img\frame + img\speed * nS_AnimationTween#
				If img\speed > 0 Then
					If img\frame > img\endf Then
						img\frame = img\endf 
						img\mode = nS_ANIM_STOP
					End If
				Else
					If img\frame < img\endf Then
						img\frame = img\endf 
						img\mode = nS_ANIM_STOP
					End If
				End If
			Case nS_ANIM_PINGPONGLOOP
				img\frame = img\frame + img\speed * nS_AnimationTween#
				If img\speed > 0 Then
					If img\frame >= img\endf + 1 Then
						img\frame = img\endf
						iTemp = img\startf
						img\startf = img\endf
						img\endf = iTemp
						img\speed = img\speed * -1
					End If
				Else
					If img\frame <= img\endf Then
						img\frame = img\endf + 0.99999
						iTemp = img\startf
						img\startf = img\endf
						img\endf = iTemp
						img\speed = img\speed * -1
					End If
				End If	
			End Select
		EntityTexture img\mesh, img\texture, Floor(img\frame)
	Else
		EntityTexture img\mesh, img\texture, frame
	End If
	
	; Lighting
	If img\bLightAffected Then
		If img\bUseVertexLighting Then
			fPosX[nS_VERTEX_BOTTOMLEFT] = x - img\x
			fPosX[nS_VERTEX_TOPLEFT] = x - img\x
			fPosX[nS_VERTEX_TOPRIGHT] = x + img\scalex * img\width - img\x
			fPosX[nS_VERTEX_BOTTOMRIGHT] = x + img\scalex * img\width - img\x
			fPosY[nS_VERTEX_BOTTOMLEFT] = y + img\scaley * img\height - img\y
			fPosY[nS_VERTEX_TOPLEFT] = y - img\y
			fPosY[nS_VERTEX_TOPRIGHT] = y - img\y
			fPosY[nS_VERTEX_BOTTOMRIGHT] = y + img\scaley * img\height - img\y
			For i = 1 To 4
				iRed = nS_AmbientRed
				iGreen = nS_AmbientGreen
				iBlue = nS_AmbientBlue
				fAlpha# = nS_AmbientAlpha
				For lights.nS_Light = Each nS_Light
					If lights\bOn Then
						If lights\bInvertedLight Then
							dir = -1
						Else
							dir = 1
						End If
						fDist# = nS_Distance(fPosX[i], fPosY[i], lights\x, lights\y)
						If fDist <= lights\iradius Then
							iRed   = iRed   + dir * img\r * ((lights\fIntensity * (Float(lights\iRed)   / 255.0) * ( 1 - (fDist / Float(lights\iRadius)))))
							iGreen = iGreen + dir * img\g * ((lights\fIntensity * (Float(lights\iGreen) / 255.0) * ( 1 - (fDist / Float(lights\iRadius)))))
							iBlue  = iBlue  + dir * img\b * ((lights\fIntensity * (Float(lights\iBlue)  / 255.0) * ( 1 - (fDist / Float(lights\iRadius)))))
							fAlpha = fAlpha + dir * img\a * (lights\fIntensity * lights\fAlpha) * ( 1 - (fDist / Float(lights\iRadius)))
						End If
					End If
				Next
				If iRed > img\r Then
					iRed = img\r
				End If
				If iGreen > img\g Then
					iGreen = img\g
				End If
				If iBlue > img\b Then
					iBlue = img\b
				End If
				If fAlpha > img\a Then
					fAlpha = img\a
				End If
				If iRed < 0 Then
					iRed = 0
				End If
				If iGreen < 0 Then
					iGreen = 0
				End If
				If iBlue < 0 Then
					iBlue = 0
				End If
				If fAlpha < 0 Then
					fAlpha = 0
				End If
				VertexColor img\surface, img\instanceID[id]+i-1, iRed, iGreen, iBlue, fAlpha
			Next
		Else
			; Not using vertex lighting
			iRed = nS_AmbientRed
			iGreen = nS_AmbientGreen
			iBlue = nS_AmbientBlue
			fAlpha# = nS_AmbientAlpha
			For lights = Each nS_Light
				If lights\bOn Then
					If lights\bInvertedLight Then
						dir = -1
					Else
						dir = 1
					End If
					fDist# = nS_Distance(x, y, lights\x, lights\y)
					If fDist <= lights\iradius Then
						iRed   = iRed   + dir * img\r * ((lights\fIntensity * (Float(lights\iRed)   / 255.0) * ( 1 - (fDist / Float(lights\iRadius)))))
						iGreen = iGreen + dir * img\g * ((lights\fIntensity * (Float(lights\iGreen) / 255.0) * ( 1 - (fDist / Float(lights\iRadius)))))
						iBlue  = iBlue  + dir * img\b * ((lights\fIntensity * (Float(lights\iBlue)  / 255.0) * ( 1 - (fDist / Float(lights\iRadius)))))
						fAlpha = fAlpha + dir * img\a * ((lights\fIntensity * lights\fAlpha                  * ( 1 - (fDist / Float(lights\iRadius)))))
					End If
				End If
			Next
			If iRed > img\r Then
				iRed = img\r
			End If
			If iGreen > img\g Then
				iGreen = img\g
			End If
			If iBlue > img\b Then
				iBlue = img\b
			End If
			If fAlpha > img\a Then
				fAlpha = img\a
			End If
			If iRed < 0 Then
				iRed = 0
			End If
			If iGreen < 0 Then
				iGreen = 0
			End If
			If iBlue < 0 Then
				iBlue = 0
			End If
			If fAlpha < 0 Then
				fAlpha = 0
			End If
			EntityColor img\mesh, iRed, iGreen, iBlue		
			EntityAlpha img\mesh, fAlpha
		End If
	End If

	; Positioning
	PositionEntity img\mesh, 0, 0, nS_ZBuffer

	; Z-Ordering + System Update
	nS_SpriteCount = nS_SpriteCount + 1
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function

; Draws a portion of the image on screen
Function nS_DrawImageRect(image%, x#, y#, rect_x#, rect_y#, rect_width#, rect_height#, frame% = 0)
	Local img.nS_Image = Object.nS_Image(image)
	Local i = 0, id = -1
	Local fx#, fy#
	Local u1#, v1#, u2#, v2#
	Local urange#, vrange#
	Local xsize#, ysize#
	Local xpos#, ypos#
	While id = -1
		If Not img\instanceUsed[i] Then
			id = i
		End If
		i = i + 1
		If i > nS_MaxInstances Then
			nS_Assert(False, "Maximum number of instances drawn for image")
			Return
		End If
	Wend
	; Calculate coordinates
	If img\centered Then
		fx# = Float(x) - 0.5 - Float(img\width) * img\scalex / 2.0 + nS_OriginX
		fy# = Float(y) - 0.5 - Float(img\height) * img\scaley / 2.0 + nS_OriginY
	Else
		fx# = Float(x) - 0.5 - Float(img\x) * img\scalex + nS_OriginX
		fy# = Float(y) - 0.5 - Float(img\y) * img\scaley + nS_OriginY
	End If
	If id > img\instanceMax Then
		img\instanceMax = id
	End If
	img\instanceUsed[id] = True
	urange = img\u2 - img\u1
	vrange = img\v2 - img\v1
	xsize = Float(rect_width) / (Float(img\width) * img\scalex)
	ysize = Float(rect_height) / (Float(img\height) * img\scaley)
	xpos = Float(rect_x) / (Float(img\width) * img\scalex)
	ypos = Float(rect_y) / (Float(img\height) * img\scaley)
	u1 = img\u1 + xpos * urange
	u2 = u1 + xsize * urange
	v1 = img\v1 + ypos * vrange
	v2 = v1 + ysize * vrange
	; The instance isn't there
	If Not(img\instanceID[id] > 0) Then
		img\instanceID[id] = AddVertex(img\surface, 0, 0, 0, 0, 0)
		AddVertex(img\surface, 0, 0, 0, 0, 0)
		AddVertex(img\surface, 0, 0, 0, 0, 0)
		AddVertex(img\surface, 0, 0, 0, 0, 0)
		AddTriangle(img\surface, img\instanceID[id], img\instanceID[id] + 3, img\instanceID[id] + 1)
		AddTriangle(img\surface, img\instanceID[id] + 2, img\instanceID[id] + 1, img\instanceID[id] + 3)
	End If
	VertexCoords img\surface, img\instanceID[id], fx, fy, nS_ZBuffer
	VertexCoords img\surface, img\instanceID[id]+1, fx, fy + rect_height, nS_ZBuffer
	VertexCoords img\surface, img\instanceID[id]+2, fx + rect_width, fy + rect_height, nS_ZBuffer
	VertexCoords img\surface, img\instanceID[id]+3, fx + rect_width, fy, nS_ZBuffer
	VertexTexCoords img\surface, img\instanceID[id], u1, v1
	VertexTexCoords img\surface, img\instanceID[id]+1, u1, v2
	VertexTexCoords img\surface, img\instanceID[id]+2, u2, v2
	VertexTexCoords img\surface, img\instanceID[id]+3, u2, v1
	
	EntityTexture img\mesh, img\texture, frame

	; Positioning
	PositionEntity img\mesh, 0, 0, nS_ZBuffer
		
	; Z-Ordering + System Update
	nS_SpriteCount = nS_SpriteCount + 1
	nS_ZBuffer = nS_ZBuffer + 0.1
End Function

; Changes the mask of an image
Function nS_MaskImage(image%, Red%, Green%, Blue%)
	Local i, x, y, size, c, r, g, b, m, iTemp
	Local img.nS_Image = Object.nS_Image(image)
	For i = 0 To img\frames - 1
		SetBuffer TextureBuffer(img\texture, i)
		LockBuffer
		For x = 0 To img\width - 1
			For y = 0 To img\height - 1
				If img\image <> 0 Then
					c = ReadPixel(x, y, ImageBuffer(img\image, i))
				Else
					c = ReadPixelFast(x, y, TextureBuffer(img\texture, i))
				End If
				r = nS_RGBR(c)
				g = nS_RGBG(c)
				b = nS_RGBB(c)
				; It is the old mask
				If r = img\maskRed And g = img\maskGreen And b = img\maskBlue Then
					WritePixelFast x, y, c
				; It is the new mask
				ElseIf r = Red And g = Green And b = Blue Then
					WritePixelFast x, y, 0
				End If
			Next
		Next
		UnlockBuffer
	Next
	img\maskRed = Red
	img\maskGreen = Green
	img\maskBlue = Blue
	If img\image <> 0 Then
		MaskImage img\image, Red, Green, Blue
	End If
	SetBuffer BackBuffer()
End Function

; Changes the handle of an image
Function nS_HandleImage(image%, x#, y#)
	Local img.nS_Image = Object.nS_Image(image)
	img\centered = False
	img\x = x
	img\y = y
	If img\image Then
		HandleImage img\image, x, y
	End If
End Function

; Sets the handle as the center of an image
Function nS_MidHandle(image%)
	Local img.nS_Image = Object.nS_Image(image)
	img\centered = True
	img\x = Float(img\width) * img\scalex / 2.0
	img\y = Float(img\height) * img\scaley / 2.0
	If img\image Then
		MidHandle img\image
	End If
End Function

; Sets the handle as the center of images automatically
Function nS_AutoMidHandle(set%)
	nS_DefaultMidHandle = set
End Function	

; Scales an image
Function nS_ScaleImage(image%, xscale#, yscale#, set% = False, updateppc% = False)
	Local img.nS_Image = Object.nS_Image(image)
	If set Then
		img\scalex = xscale
		img\scaley = yscale
	Else
		img\scalex = img\scalex * xscale
		img\scaley = img\scaley * yscale
	End If
	If img\centered Then
		img\x = Float(img\width) * img\scalex / 2.0
		img\y = Float(img\height) * img\scaley / 2.0
	End If
	If updateppc And img\image > 0 Then
		ScaleImage img\image, xscale, yscale
	End If
End Function

; Resizes an image
Function nS_ResizeImage(image%, width#, height#, updateppc% = False)
	Local img.nS_Image = Object.nS_Image(image)
	img\scalex = width / Float(img\width)
	img\scaley = height / Float(img\height)
	If img\centered Then
		img\x = Float(img\width) * img\scalex / 2.0
		img\y = Float(img\height) * img\scaley / 2.0
	End If
	If updateppc And img\image > 0 Then
		ResizeImage img\image, width, height
	End If
End Function

; Rotates an image
Function nS_RotateImage(image%, angle#, set% = False, updateppc% = False)
	Local img.nS_Image = Object.nS_Image(image)
	If set Then
		img\angle = angle
	Else
		img\angle = img\angle + angle
	End If
	If updateppc And img\image > 0 Then
		RotateImage img\image, angle
	End If
End Function

; Returns the width of an image
Function nS_ImageWidth%(image%, withscale% = True)
	Local img.nS_Image = Object.nS_Image(image)
	If withscale Then
		Return Float(img\width) * img\scalex
	Else
		Return img\width
	End If
End Function

; Returns the height of an image
Function nS_ImageHeight%(image%, withscale% = True)
	Local img.nS_Image = Object.nS_Image(image)
	If withscale Then
		Return Float(img\height) * img\scaley
	Else
		Return img\height
	End If
End Function

; Returns the image X handle
Function nS_ImageXHandle%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\x
End Function

; Returns the image Y handle
Function nS_ImageYHandle%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\y
End Function

; Checks if two images overlap
Function nS_ImagesOverlap%(image1%, x1%, y1%, image2%, x2%, y2%)
	Return RectsOverlap(x1, y1, nS_ImageWidth(image1), nS_ImageHeight(image1), x2, y2, nS_ImageWidth(image2), nS_ImageHeight(image2))
End Function

; Checks if two images collide
Function nS_ImagesCollide%(image1%, x1%, y1%, frame1%, image2%, x2%, y2%, frame2%)
	Local img1.nS_Image = Object.nS_Image(image1)
	Local img2.nS_Image = Object.nS_Image(image2)
	nS_Assert(img1\image > 0 And img2\image > 0, "One of the images doesn't have a base image. Cannot perform pixel perfect checks." )
	Return ImagesCollide(img1\image, x1, y1, frame1, img2\image, x2, y2, frame2)
End Function

; Frees an image from memory
Function nS_FreeImage(image%)
	Local img.nS_Image = Object.nS_Image(image)
	If img\image Then
		FreeImage img\image
	End If
	FreeTexture img\texture
	FreeEntity img\mesh
	Delete img
End Function

; Returns the angle of rotation of an image
Function nS_ImageAngle#(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\angle
End Function

; Creates a subimage for multiple image on same texture
Function nS_CreateSubImage%(image%, x%, y%, width%, height%)
	Local img.nS_Image = Object.nS_Image(image)
	Local newimage.nS_Image = New nS_Image
	Local w#, h#
	nS_CopyImageData(img, newimage)
	newimage\mesh = img\mesh
	newimage\surface = img\surface
	newimage\image = img\image
	newimage\texture = img\texture
	newimage\width = width
	newimage\height = height
	w = nS_ImageWidth(image)
	h = nS_ImageHeight(image)
	newimage\u1 = Float(x) / w
	newimage\v1 = Float(y) / h
	newimage\u2 = Float(x + width) / w
	newimage\v2 = Float(y + height) / h
	Return Handle(newimage)	
End Function

; Converts a Blitz image to a nSprite
Function nS_ConvertBlitzImage%(image%, framestart% = 0, frameend% = 0)
	Local w = ImageWidth(image)
	Local h = ImageHeight(image)
	Local sprite = nS_CreateImage(w, h, frameend - framestart + 1)
	Local c
	nS_DefaultImageConstructor(Object.nS_Image(sprite))
	For f = framestart To frameend
		LockBuffer nS_ImageBuffer(sprite, f - framestart)
		LockBuffer ImageBuffer(image, f)
		For x = 0 To w - 1
			For y = 0 To h - 1
				c = ReadPixelFast(x, y, ImageBuffer(image, f))
				If c <> -16777216 Then
					WritePixelFast x, y, c, nS_ImageBuffer(sprite, f - framestart)
				End If
			Next
		Next
		UnlockBuffer nS_ImageBuffer(sprite, f - framestart)
		UnlockBuffer ImageBuffer(image, f)
	Next
	Return sprite
End Function

; Duplicates an image (shared memory)
Function nS_DuplicateImage%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return nS_CreateSubImage(image, 0, 0, img\width, img\height)		
End Function

; Deletes an image handle (to remove shared memory duplicates)
Function nS_DeleteImage(image%)
	Local img.nS_Image = Object.nS_Image(image)
	For i = 0 To nS_MaxInstances
		If img\instanceUsed[i] Then
			img\instanceUsed[i] = False
			For j = 0 To 3
				VertexCoords img\surface, img\instanceID[i]+j, 0, 0, 0
			Next
		End If
	Next
	Delete img
End Function

; Flips the current image on the X and/or Y axis
Function nS_FlipImage(image%, flipx%, flipy%)
	Local img.nS_Image = Object.nS_Image(image)
	Local u2#, v2#
	u2 = img\u2
	v2 = img\v2
	If flipx Then
		img\u2 = img\u1
		img\u1 = u2
	End If
	If flipy Then
		img\v2 = img\v1
		img\v1 = v2
	End If
End Function

; Gets the texture handle
Function nS_TextureHandle%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\texture	
End Function

; Gets the mesh handle
Function nS_MeshHandle%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\mesh
End Function

; Changes the alpha of an image
Function nS_ImageAlpha(image%, alpha#)
	Local img.nS_Image = Object.nS_Image(image)
	img\a = alpha
	EntityAlpha img\mesh, alpha
End Function

; Changes the color of an image
Function nS_ImageColor(image%, red%, green%, blue%)
	Local img.nS_Image = Object.nS_Image(image)
	img\r = red
	img\g = green
	img\b = blue
	EntityColor img\mesh, red, green, blue
End Function

; Sets the entityFX for the image's mesh
Function nS_ImageFX(image%, fx%)
	Local img.nS_Image = Object.nS_Image(image)
	img\fx = fx
	EntityFX img\mesh, fx
End Function

; Sets the entity blend mode for the image's mesh
Function nS_ImageBlend(image%, blend%)
	Local img.nS_Image = Object.nS_Image(image)
	img\blend = blend
	EntityBlend img\mesh, blend
End Function

; Changes the color of one of the vertex
Function nS_ImageVertexColor(image%, vertexid%, red%, green%, blue%, alpha# = 1)
	Local img.nS_Image = Object.nS_Image(image)
	If Not((nS_FX_FORCEALPHABLEND + nS_FX_VERTEXCOLOR) And img\fx) Then
		nS_Warning("Image isn't set for vertex coloring and forced alpha blend and is assigned vertex color - " + img\name)
	End If
	img\vertexcolored[vertexid] = True
	img\vertexr[vertexid] = red
	img\vertexg[vertexid] = green
	img\vertexb[vertexid] = blue
	img\vertexa[vertexid] = alpha
End Function

; Changes the position of one of the vertex
Function nS_ImageVertexPosition(image%, vertexid%, xoff#, yoff#)
	Local img.nS_Image = Object.nS_Image(image)
	img\vertox[vertexid] = xoff
	img\vertoy[vertexid] = yoff
End Function

; Automatically animate an image
Function nS_AnimateImage(image%, mode%, speed#, startframe#, endframe#)
	img.nS_Image = Object.nS_Image(Image)
	img\mode = mode
	img\speed# = speed#
	img\startf# = startframe#
	img\endf# = endframe#
	img\frame# = startframe#
End Function

; Returns TRUE if the anim is currently playing
Function nS_IsAnimImagePlaying%(image%)
	img.nS_Image = Object.nS_Image(image)
	Return Not(img\mode = nS_ANIM_STOP)
End Function

; Returns the current animation time of the image
Function nS_GetAnimTime#(image%)
	img.nS_Image = Object.nS_Image(image)
	Return img\frame
End Function

; Sets the animation tween that will be used
Function nS_SetAnimTween(tween# = 1)
	nS_AnimationTween = tween
End Function

; Returns the alpha of an image
Function nS_GetImageAlpha#(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\a
End Function

; Returns the red level
Function nS_GetImageRed%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\r
End Function

; Returns the green level
Function nS_GetImageGreen%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\g
End Function

; Returns the blue level
Function nS_GetImageBlue%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\b
End Function

; Returns the FX set for the image's entity
Function nS_GetImageFX%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\fx
End Function

; Returns the entity blend mode set
Function nS_GetImageBlend%(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\blend
End Function

; Returns the image scale in X
Function nS_ImageScaleX#(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\scalex
End Function

; Returns the image scale in Y
Function nS_ImageScaleY#(image%)
	Local img.nS_Image = Object.nS_Image(image)
	Return img\scaley
End Function