; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Globals and Constants
; ======================================================================

Global nS_AmbientRed = 0
Global nS_AmbientGreen = 0
Global nS_AmbientBlue = 0
Global nS_AmbientAlpha# = 1

; ======================================================================
; Different objects
; ======================================================================

; Lights
Type nS_Light
	Field X#				; X Position
	Field Y#				; Y Position
	Field iRadius#			; Light Radius (every outside isn't counted)
	Field iRed				; Red level
	Field iGreen			; Green level
	Field iBlue				; Blue level
	Field fAlpha#			; Alpha level
	Field fIntensity#		; Intensity level (% it affects)
	Field bOn				; Is the light on or off?
	Field bInvertedLight    ; Is it an inverted light?
End Type

; ======================================================================
; Light Functions for Images
; ======================================================================

; Sets the light affect of an image
Function nS_ImageLightAffect(image%, bAffectedByLight%, bUseVertexLighting% = False)
	Local sprite.nS_Image = Object.nS_Image(Image)
	sprite\bLightAffected = bAffectedByLight
	sprite\bUseVertexLighting = bUseVertexLighting
End Function

; ----------------------------------------------------------------------

; Gets if the image is light affected or not
Function nS_IsImageLightAffected%(image%)
	Local sprite.nS_Image = Object.nS_Image(Image)
	Return sprite\bLightAffected
End Function

; ======================================================================
; Light Functions
; ======================================================================

; ----------------------------------------------------------------------

; Sets the ambient light level (default 0, 0, 0, 1)
Function nS_AmbientLight(red% = 0, green% = 0, blue% = 0, alpha# = 1)
	nS_AmbientRed = red
	nS_AmbientGreen = green
	nS_AmbientBlue = blue
	nS_AmbientAlpha = alpha
End Function

; Creates a light entity
Function nS_CreateLight%(x%, y%, iRadius%, iRed%, iGreen%, iBlue%, fIntensity#, fAlpha# = 0, bInvertedLight% = False)
	Local light.nS_Light = New nS_Light
	light\x = x
	light\y = y
	light\iRadius = iRadius
	light\iRed = iRed
	light\iGreen = iGreen
	light\iBlue = iBlue
	light\fIntensity = fIntensity
	light\fAlpha = fAlpha
	light\bOn = True
	light\bInvertedLight = bInvertedLight
	Return Handle(light)
End Function

; ----------------------------------------------------------------------

; Moves a light entity
Function nS_MoveLight(light%, x#, y#)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	lights\x = x
	lights\y = y
End Function

; ----------------------------------------------------------------------

; Changes the parameters of a light entity
Function nS_ChangeLight(light%, fRadius# = -1, iRed% = -1, iGreen% = -1, iBlue% = -1, fIntensity# = -1, fAlpha# = -1, bInvertedLight% = -1 )
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	If iRadius <> -1 Then
		lights\iRadius = fRadius
	End If
	If iRed <> -1 Then
		lights\iRed = iRed
	End If
	If iGreen <> -1 Then
		lights\iGreen = iGreen
	End If
	If iBlue <> -1 Then
		lights\iBlue = iBlue
	End If
	If fAlpha <> -1 Then
		lights\fAlpha = fAlpha
	End If
	If fIntensity <> -1 Then
		lights\fIntensity = fIntensity
	End If
	If bInvertedLight <> -1 Then
		lights\bInvertedLight = bInvertedLight
	End If
End Function

; ----------------------------------------------------------------------

; Frees a light from memory
Function nS_FreeLight(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Delete lights
End Function

; ----------------------------------------------------------------------

; Turns the light on or off
Function nS_SwitchLight(light%, bOn = -1)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	If bOn = -1 Then
		If lights\bOn Then
			lights\bOn = False
		Else
			lights\bOn = True
		End If
	Else
		lights\bOn = bOn
	End If
End Function

; ----------------------------------------------------------------------

; Copies a light entity and returns the new handle
Function nS_CopyLight%(light%)
	Local lights.nS_Light, newLight.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	newlight = New nS_Light
	newlight\x = lights\x
	newlight\y = lights\y
	newlight\iRadius = lights\iRadius
	newlight\iRed = lights\iRed
	newlight\iGreen = lights\iGreen
	newlight\iBlue = lights\iBlue
	newlight\fAlpha = lights\fAlpha
	newlight\fIntensity = lights\fIntensity
	newlight\bOn = lights\bOn
	Return Handle(newlight)
End Function

; ----------------------------------------------------------------------

; Gets the light x position
Function nS_GetLightX#(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
 	Return lights\x
End Function

; ----------------------------------------------------------------------

; Gets the light y position
Function nS_GetLightY#(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\y
End Function

; ----------------------------------------------------------------------

; Gets the light radius
Function nS_GetLightRadius#(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\iRadius
End Function

; ----------------------------------------------------------------------

; Gets the light red level
Function nS_GetLightRed%(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\iRed
End Function

; ----------------------------------------------------------------------

; Gets the light green level
Function nS_GetLightGreen%(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\iGreen
End Function

; ----------------------------------------------------------------------

; Gets the light blue level
Function nS_GetLightBlue%(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\iBlue
End Function

; ----------------------------------------------------------------------

; Gets the light alpha level
Function nS_GetLightAlpha#(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\fAlpha
End Function

; ----------------------------------------------------------------------

; Gets the light intensity
Function nS_GetLightIntensity#(light%)
	Local lights.nS_Light
	lights.nS_Light = Object.nS_Light(Light)
	Return lights\fIntensity
End Function