; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Different objects
; ======================================================================

; Emitter Object
Type nS_Emitter
	Field iX					; X Position
	Field iY					; Y Position
	Field iCount				; Particle Count
	Field bExists				; Emitter still exists?
	; Emitter Settings
	Field iLife					; Emitter Life in particle count
	Field fFlow#				; Particles per second
	Field fFlowTime#			; Flow time for new particles
	; Particle Settings
	Field sprParticle			; Particle Sprite
	Field fParticleLife#		; Particle Life
	Field fLifeSpread#			; Life spread for particle
	; Velocity Settings
	Field fVelX#				; X Velocity
	Field fVelY#				; Y Velocity
	Field fVelSpreadX#			; X Velocity Spread
	Field fVelSpreadY#			; Y Velocity Spread
	Field fGravity#				; Gravity
	Field fAccX#				; X Velocity Acceleration
	Field fAccY#				; Y Velocity Acceleration
	; Angle Settings
	Field fAngle#				; Initial Angle
	Field fAngleSpread#			;  The spread
	Field fRotation#			; Rotation of the particle
	Field fRotationSpread#		;  The spread
	; Scale Settings
	Field fScaleX#				; X Scale of the particle
	Field fScaleY#				; Y Scale of the particle
	Field fScaleXSpread#		; The spread in X
	Field fScaleYSpread#		; The spread in Y
	Field fGrowX#				; The growth in X
	Field fGrowY#				; The growth in Y
End Type

; Particle Object
Type nS_Particle
	Field oEmitter.nS_Emitter	; Emitter Owner
	Field sprImage				; The sprite
	Field fLife#				; Remaining Life
	Field fX#					; X Position
	Field fY#					; Y Position
	Field fVelX#				; X Velocity
	Field fVelY#				; Y Velocity
	Field fAngle#				; Angle
	Field fRotation#			; Rotation
	Field fScaleX#				; Scale in X
	Field fScaleY#				; Scale in Y
	Field fAlpha#				; Alpha level
End Type

; ======================================================================
; Functions
; ======================================================================

; Creates a new emitter and return its handle
Function nS_CreateEmitter%()
	Local newEmitter.nS_Emitter = New nS_Emitter
	; Apply some default values
	newEmitter\fScaleX# = 1
	newEmitter\fScaleY# = 1
	newEmitter\fGrowX# = 1
	newEmitter\fGrowY# = 1
	newEmitter\bExists = True
	Return Handle(newEmitter)
End Function

; ----------------------------------------------------------------------

; Sets the basic emitter settings
Function nS_SetEmitter(ID%, iLife, fFlow#)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\iLife = iLife
	emitter\fFlow# = 1.0 / fFlow# ; s per particle
End Function

; ----------------------------------------------------------------------

; Sets the particle settings
Function nS_SetParticle(ID%, sprPart%, fLife#, fLifeSpread#)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\sprParticle = sprPart
	emitter\fParticleLife = fLife
	emitter\fLifeSpread = fLifeSpread
End Function

; ----------------------------------------------------------------------

; Sets the velocity settings
Function nS_SetVelocity(ID%, fVelX#, fVelY#, fVelSpreadX#, fVelSpreadY#, fAccX#, fAccY#, fGravity#)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\fVelX# = fVelX#
	emitter\fVelY# = fVelY#
	emitter\fVelSpreadX# = fVelSpreadX#
	emitter\fVelSpreadY# = fVelSpreadY#
	emitter\fAccX# = fAccX#
	emitter\fAccY# = fAccY#
	emitter\fGravity# = fGravity#
End Function

; ----------------------------------------------------------------------

; Sets the angle settings
Function nS_SetRotation(ID%, fAngle#, fAngleSpread#, fRotation#, fRotationSpread#)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\fAngle# = fAngle#
	emitter\fAngleSpread# = fAngleSpread#
	emitter\fRotation# = fRotation#
	emitter\fRotationSpread# = fRotationSpread#
End Function

; ----------------------------------------------------------------------

; Sets the scale settings
Function nS_SetScale(ID%, fScaleX#, fScaleY#, fScaleXSpread#, fScaleYSpread#, fGrowX#, fGrowY#)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\fScaleX# = fScaleX#
	emitter\fScaleY# = fScaleY#
	emitter\fScaleXSpread# = fScaleXSpread#
	emitter\fScaleYSpread# = fScaleYSpread#
	emitter\fGrowX# = fGrowX#
	emitter\fGrowY# = fGrowY#
End Function

; ----------------------------------------------------------------------

; Positions the emitter
Function nS_PositionEmitter(ID%, iX%, iY%)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\iX = iX
	emitter\iY = iY
End Function

; ----------------------------------------------------------------------

; Moves an emitter and particles with it (if needed)
Function nS_MoveEmitter(ID%, iX%, iY%, bAffectParticle% = True)
	Local emitter.nS_Emitter
	Local particle.nS_Particle
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\iX = emitter\iX + iX
	emitter\iY = emitter\iY + iY
	If bAffectParticle Then
		For particle = Each nS_Particle
			If particle\oEmitter = emitter Then
				particle\fX = particle\fX + iX
				particle\fY = particle\fY + iY
			End If
		Next
	End If
End Function

; ----------------------------------------------------------------------

; Returns true if the emitter exists
Function nS_EmitterExists%(ID%)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	Return emitter\bExists
End Function

; ----------------------------------------------------------------------

; Kills the particles of an emitter
Function nS_KillParticles(ID%)
	Local particle.nS_Particle
	For particle = Each nS_Particle
		If particle\oEmitter = Object.nS_Emitter(ID) Then
			Delete particle
		End If
	Next
End Function

; ----------------------------------------------------------------------

; Kills an emitter before it ends emitting
Function nS_KillEmitter(ID%)
	Local emitter.nS_Emitter
	emitter.nS_Emitter= Object.nS_Emitter(ID)
	emitter\bExists = False
End Function

; ----------------------------------------------------------------------

; Clears all particles
Function nS_ClearParticles()
	Local particle.nS_Particle
	For particle = Each nS_Particle
		Delete particle
	Next
End Function

; ----------------------------------------------------------------------

; Clears all emitters
Function nS_ClearEmitters()
	Local emitter.nS_Emitter
	For emitter = Each nS_Emitter
		emitter\bExists = False
	Next
End Function

; ----------------------------------------------------------------------

; Updates all emitters (launch new particles, etc)
Function nS_UpdateEmitters(fFrameTweening# = 1)
	Local emitter.nS_Emitter
	Local newParts
	For emitter = Each nS_Emitter
		; Check if emitter still exists
		If emitter\bExists Then
			; Check if its time for a new particle
			emitter\fFlowTime = emitter\fFlowTime + fFrameTweening#
			while(emitter\fFlowTime >= emitter\fFlow)
				emitter\fFlowTime = emitter\fFlowTime - emitter\fFlow
				nS_GenerateParticle(emitter)
				; Life handling
				If emitter\iLife > 0 Then
					emitter\iLife = emitter\iLife - 1
					If emitter\iLife <= 0 Then
						emitter\bExists = False
						emitter\fFlowTime = 0
					End If
				End If
			wend
			
			; Reset particle count
			emitter\iCount = 0
		Else
			; Delete if no more live particles
			If emitter\iCount = 0 Then
				Delete emitter
			End If
		End If
	Next
End Function

; ----------------------------------------------------------------------

; Updates all particles (velocity, life, draw, etc)
Function nS_UpdateParticles%(fFrameTweening# = 1)
	Local particle.nS_Particle
	Local iTotal
	Local fNewScaleX#, fNewScaleY#
	For particle = Each nS_Particle
		If particle\fLife > 0 Then
			; Image update
			nS_RotateImage(particle\sprImage, particle\fRotation# * fFrameTweening#)
			nS_ScaleImage(particle\sprImage, particle\fScaleX#, particle\fScaleY#, True)
			nS_DrawImage(particle\sprImage, particle\fX#, particle\fY#)
			; Count Updates
			iTotal = iTotal + 1
			particle\oEmitter\iCount = particle\oEmitter\iCount + 1
			; Position Update
			particle\fX# = particle\fX# + particle\fVelX# * fFrameTweening#
			particle\fY# = particle\fY# + particle\fVelY# * fFrameTweening#
			; Velocity Update
			fNewScaleX# = (particle\oEmitter\fAccX# - 1) * fFrameTweening# + 1
			fNewScaleY# = (particle\oEmitter\fAccY# - 1) * fFrameTweening# + 1
			particle\fVelX# = particle\fVelX# * fNewScaleX#
			particle\fVelY# = (particle\fVelY# * fNewScaleY#) + particle\oEmitter\fGravity# * fFrameTweening#
			; Scale Update
			fNewScaleX# = (particle\oEmitter\fGrowX# - 1) * fFrameTweening# + 1
			fNewScaleY# = (particle\oEmitter\fGrowY# - 1) * fFrameTweening# + 1
			particle\fScaleX# = particle\fScaleX# * fNewScaleX#
			particle\fScaleY# = particle\fScaleY# * fNewScaleY#
			; Life Update
			particle\fLife = particle\fLife - fFrameTweening#
		End If
	Next
	Return iTotal
End Function

; ----------------------------------------------------------------------

; Creates a new particle from emitter
Function nS_GenerateParticle(emitter.nS_Emitter)
	Local particle.nS_Particle
	For particle = Each nS_Particle
		If particle\fLife < 0 And emitter = particle\oEmitter Then
			found = True
			Exit
		End If
	Next
	If Not found
		particle = New nS_Particle
		particle\sprImage = nS_DuplicateImage(emitter\sprParticle)
	End If
	particle\oEmitter = emitter
	particle\fX = emitter\iX
	particle\fY = emitter\iY
	particle\fVelX# = emitter\fVelX# + nS_SpreadRnd(emitter\fVelSpreadX#)
	particle\fVelY# = emitter\fVelY# + nS_SpreadRnd(emitter\fVelSpreadY#)
	particle\fLife = emitter\fParticleLife + nS_SpreadRnd(emitter\fLifeSpread)
	particle\fAngle# = emitter\fAngle# + nS_SpreadRnd(emitter\fAngleSpread#)
	particle\fRotation# = emitter\fRotation# + nS_SpreadRnd(emitter\fRotationSpread#)
	particle\fScaleX# = emitter\fScaleX# + nS_SpreadRnd(emitter\fScaleXSpread#)
	particle\fScaleY# = emitter\fScaleY# + nS_SpreadRnd(emitter\fScaleYSpread#)
	nS_RotateImage(particle\sprImage, particle\fAngle#)
End Function

; ======================================================================
; Functions needed by the lib
; ======================================================================
Function nS_SpreadRnd#(fValue#)
	Return (Rnd(0, fValue# * 2) - fValue#)
End Function