; nSprite 2 - 3D Accelerated 2D Games Creation Package for Blitz3D
; Last Updated: November 15th 2020
; Version: 1.2.0

; Copyright (c) 2020 Jocelyn Perreault
;
; This software is provided 'as-is', without any express or implied
; warranty. In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.

; ======================================================================
; Globals and Constants
; ======================================================================

Const nS_CurrentVersion = 120
Const nS_MaxInstances = 512

Const nS_Debug = False

Global nS_ZBuffer#

Global nS_AspectRatio#
Global nS_Camera

Global nS_pvtTopLeft
Global nS_Scale#

Global nS_imgBlank

Global nS_OriginX
Global nS_OriginY

Global nS_SpriteCount

Global nS_DefaultMidHandle
Global nS_DefaultFX

Global nS_AnimationTween#

; Base Shapes
Const nS_SHAPE_FILLRECT_LINE_OR_POINT = 1
Const nS_SHAPE_HOLLOWRECT = 2
Const nS_SHAPE_FILLOVAL = 3
Const nS_SHAPE_HOLLOWOVAL = 4
Const nS_SHAPE_TRI = 5
Const nS_MAXSHAPES = 512
Global nS_ColorRed = 255
Global nS_ColorGreen = 255
Global nS_ColorBlue = 255
Global nS_ColorAlpha# = 1.0
Global nS_ShapesMesh
Global nS_ShapesSurface
Global nS_ShapesUsed[nS_MAXSHAPES]
Global nS_ShapesStart[nS_MAXSHAPES]
Global nS_ShapesType[nS_MAXSHAPES]
Global nS_ShapesVertices[nS_MAXSHAPES]
Global nS_ShapesCount = 0

Const nS_FX_NONE = 0
Const nS_FX_FULLBRIGHT = 1
Const nS_FX_VERTEXCOLOR = 2
Const nS_FX_FLATSHADED = 4
Const nS_FX_DISABLEFOG = 8
Const nS_FX_DISABLEBACKFACECULLING = 16
Const nS_FX_FORCEALPHABLEND = 32

Const nS_BLEND_ALPHA = 1
Const nS_BLEND_MULTIPLY = 2
Const nS_BLEND_ADDITIVE = 3

Const nS_VERTEX_TOPLEFT = 1
Const nS_VERTEX_BOTTOMLEFT = 2
Const nS_VERTEX_BOTTOMRIGHT = 3
Const nS_VERTEX_TOPRIGHT = 4

Const nS_ANIM_STOP = 0
Const nS_ANIM_LOOP = 1
Const nS_ANIM_PINGPONG = 2
Const nS_ANIM_ONESHOT = 3
Const nS_ANIM_PINGPONGLOOP = 4

; ======================================================================
; System Functions
; ======================================================================

; Logs an error in debug and gives runtime error in non-debug
Function nS_Assert(condition, msg$)
	If Not condition Then
		msg = "nSprite 2 Error - " + msg
		If nS_Debug Then
			DebugLog msg
		Else
			RuntimeError msg
		End If
	End If
End Function

; Sends a warning in the debug log
Function nS_Warning(msg$)
	msg = "nSprite 2 Warning - " + msg
	DebugLog msg
End Function

; Find the next power of 2 texture size for a rectangle
Function nS_GetTextureSize%(width, height)
	Local size = width
	If height > size Then
		size = height
	End If
	size = Ceil(Log(size) / Log(2))
	Return 2 ^ size
End Function

; Returns the RGB code from individual values
Function nS_RGB%(R, G, B)
	Return ((R Shl 16) Or (G Shl 8) Or B)
End Function

; Copies the image data from one to another
Function nS_CopyImageData(image.nS_Image, newimage.nS_Image)
	newimage\width = image\width
	newimage\height = image\height
	newimage\name = image\name
	newimage\flags = image\flags
	newimage\frames = image\frames
	newimage\size = image\size
	newimage\u1 = image\u1
	newimage\v1 = image\v1
	newimage\u2 = image\u2
	newimage\v2 = image\v2
	newimage\scalex = image\scalex
	newimage\scaley = image\scaley
	newimage\angle = image\angle
	newimage\x = image\x
	newimage\y = image\y
	newimage\centered = image\centered
	newimage\MaskRed = image\MaskRed
	newimage\MaskGreen = image\MaskGreen
	newimage\MaskBlue = image\MaskBlue
	newimage\r = image\r
	newimage\g = image\g
	newimage\b = image\b
	newimage\a = image\a
	newimage\fx = image\fx
	newimage\blend = image\blend
	For i = 1 To 4
		newimage\vertexcolored[i] = image\vertexcolored[i]
		newimage\vertexr[i] = image\vertexr[i]
		newimage\vertexg[i] = image\vertexg[i]
		newimage\vertexb[i] = image\vertexb[i]
		newimage\vertexa[i] = image\vertexa[i]
	Next
	newimage\mode = image\mode
	newimage\speed = image\speed
	newimage\startf = image\startf
	newimage\endf = image\endf
	newimage\frame = image\frame
	newimage\bLightAffected = image\bLightAffected
	newimage\bUseVertexLighting = image\bUseVertexLighting
	nS_CopyCollisions(Handle(image), Handle(newimage))
End Function

; Default image constructor functions
Function nS_DefaultImageConstructor(image.nS_Image)
	Local img = Handle(image)
	image\mesh = CreateMesh(nS_pvtTopLeft)
	image\surface = CreateSurface(image\mesh)
	AddVertex(image\surface, 0, 0, 0)
	
	EntityTexture image\mesh, image\texture, 0
	
	If nS_DefaultMidHandle Then
		nS_MidHandle(img)
	End If
	nS_ImageFX(img, nS_DefaultFX)
	
	image\scalex = 1
	image\scaley = 1
	image\r = 255
	image\g = 255
	image\b = 255
	image\a = 1
End Function

; Gets the red value from RGB
Function nS_RGBR%(RGB)
	Return ((RGB Shr 16) And 255)
End Function

; Gets the green value from RGB
Function nS_RGBG%(RGB)
	Return ((RGB Shr 8) And 255)
End Function

; Gets the blue value from RGB
Function nS_RGBB%(RGB)
	Return (RGB And 255)
End Function

; Returns the log in base 2
Function nS_Log2#(x#) 
	Return (Log(x) / Log(2))
End Function 

; Gets the distance between two points
Function nS_Distance#(x1#, y1#, x2#, y2#)
	Local width# = x1# - x2#
	Local height# = y1# - y2#
	Local dist# = Sqr#(width * width + height * height) 
	Return dist
End Function

; Find a shape position in the array or create it if needed and then colorize it properly
Function nS_FindShape%(iShapeType%, iVertices%)
	Local i%, p% = 0
	; Find spot
	For i = 1 To nS_ShapesCount
		If Not nS_ShapesUsed[i] Then
			If nS_ShapesType[i] = iShapeType Then
				If nS_ShapesVertices[i] = iVertices Then
					nS_ShapesUsed[i] = True
					p = i
					Exit
				End If
			End If
		End If		
	Next
	; If not, create new one
	If p = 0 Then
		nS_ShapesCount = nS_ShapesCount + 1
		nS_Assert(nS_ShapesCount <= nS_MAXSHAPES, "Too many basic shapes drawn on screen!")
		p = nS_ShapesCount
		nS_ShapesUsed[p] = True
		nS_ShapesVertices[p] = iVertices
		nS_ShapesType[p] = iShapeType
		nS_ShapesStart[p] = AddVertex(nS_ShapesSurface, 0, 0, 0)
		For i = 2 To iVertices
			AddVertex(nS_ShapesSurface, 0, 0, 0)
		Next
		; Add triangles
		Select iShapeType
			Case nS_SHAPE_FILLRECT_LINE_OR_POINT
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+2, nS_ShapesStart[p]+3
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+3, nS_ShapesStart[p]+1
			Case nS_SHAPE_HOLLOWRECT
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+0, nS_ShapesStart[p]+2, nS_ShapesStart[p]+4
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+4, nS_ShapesStart[p]+2, nS_ShapesStart[p]+6
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+0, nS_ShapesStart[p]+5, nS_ShapesStart[p]+1
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+0, nS_ShapesStart[p]+4, nS_ShapesStart[p]+5
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+6, nS_ShapesStart[p]+2, nS_ShapesStart[p]+3
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+6, nS_ShapesStart[p]+3, nS_ShapesStart[p]+7
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+5, nS_ShapesStart[p]+7, nS_ShapesStart[p]+1
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+1, nS_ShapesStart[p]+7, nS_ShapesStart[p]+3
			Case nS_SHAPE_FILLOVAL
				For i = 1 To iVertices - 2
					AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+i+1, nS_ShapesStart[p]+i
				Next
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+1, nS_ShapesStart[p]+iVertices-1
			Case nS_SHAPE_HOLLOWOVAL
				For i = 0 To iVertices - 4 Step 2
					AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+i+2, nS_ShapesStart[p]+i+1, nS_ShapesStart[p]+i
					AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+i+2, nS_ShapesStart[p]+i+3, nS_ShapesStart[p]+i+1
				Next
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+iVertices-1, nS_ShapesStart[p]+iVertices-2
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+1, nS_ShapesStart[p]+iVertices-1
			Case nS_SHAPE_TRI
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p], nS_ShapesStart[p]+1, nS_ShapesStart[p]+2
				AddTriangle nS_ShapesSurface, nS_ShapesStart[p]+2, nS_ShapesStart[p]+1, nS_ShapesStart[p]
		End Select
	End If
	; Colorize it
	For i = nS_ShapesStart[p] To nS_ShapesStart[p] + iVertices - 1
		VertexColor nS_ShapesSurface, i, nS_ColorRed, nS_ColorGreen, nS_ColorBlue, nS_ColorAlpha
	Next
	Return p
End Function